from collections.abc import Sequence
from dataclasses import field, make_dataclass
from typing import Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Sym, SymConfig
from ravenpy.config.commands import (
    HRU,
    PL,
    LandUseClasses,
    LandUseParameterList,
    SoilClasses,
    SoilParameterList,
    SoilProfiles,
    VegetationClasses,
)
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


P = dataclass(
    make_dataclass(
        "Params",
        [(f"X{i:02}", Sym, field(default=Variable(f"X{i:02}"))) for i in range(1, 22)],
    ),
    config=SymConfig,
)

# Bug: RavenC tries to write two variables with the same name `Snow Melt (Liquid) [mm]` to netCDF.


class LandHRU(HRU):
    land_use_class: str = "OPEN_1"
    veg_class: str = "FOREST"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[LandHRU]


class HYPR(Config):
    """
    The HYdrological model for Prairie Region (HYPR) is based on the
    conceptual Hydrologiska Byrans Vattenbalansavdelning (HBV)-light
    model. HBV is modified to work in the prairies by incorporating a
    conceptual lateral-flow component to represent the pothole storage
    complexities. HYPR can be used for prairie streamflow simulation.

    References
    ----------
    Mohamed I. Ahmed, Amin Elshorbagy, and Alain Pietroniro (2020).
    Toward Simple Modeling Practices in the Complex Canadian Prairie
    Watersheds. Journal of Hydrologic Engineering, 25(6), 04020024.
    doi: 10.1061/(ASCE)HE.1943-5584.0001922.
    """

    # TRANSLATION OF HYPR PARAMETERS
    # -------------------------------------------------------
    # HYPR --> RAVEN
    # ................
    # TT              --> DD_MELT_TEMP                                         x01
    # TT              --> RAINSNOW_TEMP                                        x02
    # CWH             --> SNOW_SWI                                             x03
    # LP              --> FIELD_CAPACITY                                       x04
    # K0              --> BASEFLOW_COEFF[2]                                    x05
    # K1              --> BASEFLOW_COEFF[1]                                    x06
    # B               --> PDMROF_B                                             x07
    # MAXPA           --> MAX_DEP_AREA_FRAC                                    x08
    # PWR             --> PONDED_EXP = 2/PWR                                   x09
    # UZL             --> Threshold storage = STORAGE_THRESHOLD               x10
    # FC              --> max storage capacity  = THICKNESS[0]*POROSITY[0]     x11
    # CRFR            --> REFREEZE_FACTOR                                      x12
    # MRF             --> HBV_MELT_FOR_CORR                                    x13
    # CMAX            --> DEP_MAX = CMAX/(B+1)                                 x14
    # MAXBAS          --> TIME_CONC                                            x15
    # BETA            --> HBV_BETA                                             x16
    # C0=KMIN         --> MELT_FACTOR                                          x17
    # C0=KMIN         --> MIN_MELT_FACTOR                                      x18
    # TCALT           --> AdiabaticLapseRate (from HBV)                        x19
    # PCALT           --> PRECIP_LAPSE       (from HBV)                        x20
    # SCF             --> :SnowCorrection                                      x21

    params: P = P()
    hrus: HRUs = Field(
        [
            LandHRU(),
        ],
        alias="HRUs",
    )
    netcdf_attribute: dict[str, str] = {"model_id": "HYPR"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_TRI_CONVOLUTION", alias="CatchmentRouting")
    evaporation: o.Evaporation = Field(o.Evaporation.FROMMONTHLY, alias="Evaporation")
    ow_evaporation: o.Evaporation = Field(o.Evaporation.FROMMONTHLY, alias="OW_Evaporation")
    rain_snow_fraction: o.RainSnowFraction = Field(o.RainSnowFraction.HBV, alias="RainSnowFraction")
    sw_radiation_method: o.SWRadiationMethod = Field(o.SWRadiationMethod.DEFAULT, alias="SWRadiationMethod")
    sw_cloud_correct: o.SWCloudCorrect = Field(o.SWCloudCorrect.NONE, alias="SWCloudCorrect")
    sw_canopy_correct: o.SWCanopyCorrect = Field(o.SWCanopyCorrect.NONE, alias="SWCanopyCorrect")
    lw_radiation_method: o.LWRadiationMethod = Field(o.LWRadiationMethod.DEFAULT, alias="LWRadiationMethod")
    potential_melt_method: o.PotentialMeltMethod = Field(o.PotentialMeltMethod.HBV, alias="PotentialMeltMethod")
    cloud_cover_method: o.CloudCoverMethod = Field(o.CloudCoverMethod.NONE, alias="CloudCoverMethod")
    precip_icept_frac: o.PrecipIceptFract = Field(o.PrecipIceptFract.USER, alias="PrecipIceptFrac")
    soil_model: rc.SoilModel = Field(3, alias="SoilModel")

    hydrologic_processes: Sequence[Union[rc.Process, p.Conditional]] = Field(
        [
            p.SnowRefreeze(algo="FREEZE_DEGREE_DAY", source="SNOW_LIQ", to="SNOW"),
            p.Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.CanopyEvaporation(algo="CANEVP_ALL", source="CANOPY", to="ATMOSPHERE"),
            p.CanopySublimation(algo="CANEVP_ALL", source="CANOPY_SNOW", to="ATMOSPHERE"),
            p.SnowBalance(algo="SNOBAL_SIMPLE_MELT", source="SNOW", to="PONDED_WATER"),
            p.Infiltration(algo="INF_HBV", source="PONDED_WATER", to="MULTIPLE"),
            p.Flush(algo="RAVEN_DEFAULT", source="SURFACE_WATER", to="PONDED_WATER"),
            p.Abstraction(algo="ABST_PDMROF", source="PONDED_WATER", to="DEPRESSION"),
            p.Flush(algo="RAVEN_DEFAULT", source="SURFACE_WATER", to="SOIL[1]"),
            p.SoilEvaporation(algo="SOILEVAP_HYPR", source="MULTIPLE", to="ATMOSPHERE"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[1]", to="SURFACE_WATER"),
            p.Baseflow(algo="BASE_THRESH_STOR", source="SOIL[1]", to="SURFACE_WATER"),
        ],
        alias="HydrologicProcesses",
    )
    lake_storage: o.StateVariables = Field("SOIL[2]", alias="LakeStorage")

    soil_classes: SoilClasses = Field(
        [{"name": "TOPSOIL"}, {"name": "SLOW_RES"}, {"name": "FAST_RES"}],
        alias="SoilClasses",
    )
    vegetation_classes: VegetationClasses = Field(
        [{"name": "FOREST", "max_ht": 0, "max_lai": 0, "max_leaf_cond": 1e99}],
        alias="VegetationClasses",
    )
    land_use_classes: LandUseClasses = Field(
        [rc.LU(name="OPEN_1", impermeable_frac=0.0, forest_coverage=0.0)],
        alias="LandUseClasses",
    )

    soil_profiles: SoilProfiles = Field(
        [
            {
                "name": "DEFAULT_P",
                "soil_classes": ("TOPSOIL", "FAST_RES", "SLOW_RES"),
                "thicknesses": (P.X11, 1e99, 1e99),
            },
        ],
        alias="SoilProfiles",
    )

    global_parameter: dict = Field(
        {
            "RAINSNOW_TEMP": P.X02,
            "RAINSNOW_DELTA": 0,
            "SNOW_SWI": P.X03,
            "AdiabaticLapseRate": P.X19,
            "PrecipLapse": P.X20,
        }
    )
    soil_parameter_list: SoilParameterList = Field(
        {
            "parameters": (
                "POROSITY",
                "FIELD_CAPACITY",
                "SAT_WILT",
                "HBV_BETA",
                "MAX_CAP_RISE_RATE",
                "MAX_PERC_RATE",
                "BASEFLOW_COEFF",
                "BASEFLOW_N",
                "BASEFLOW_COEFF2",
                "STORAGE_THRESHOLD",
            ),
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(1.0, P.X04, 0.0, P.X16, 0, 0, 0, 0, 0, 0),
                ),
                PL(
                    name="FAST_RES",
                    values=(
                        "_DEFAULT",
                        "_DEFAULT",
                        0.0,
                        "_DEFAULT",
                        "_DEFAULT",
                        0,
                        P.X06,
                        1.0,
                        P.X05,
                        P.X10,
                    ),
                ),
                PL(
                    name="SLOW_RES",
                    values=(
                        "_DEFAULT",
                        "_DEFAULT",
                        0.0,
                        "_DEFAULT",
                        "_DEFAULT",
                        "_DEFAULT",
                        0.01,
                        1.0,
                        0.05,
                        0,
                    ),
                ),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: LandUseParameterList = Field(
        {
            "parameters": (
                "MELT_FACTOR",
                "MIN_MELT_FACTOR",
                "HBV_MELT_FOR_CORR",
                "REFREEZE_FACTOR",
                "HBV_MELT_ASP_CORR",
                "DD_MELT_TEMP",
                "FOREST_COVERAGE",
                "PONDED_EXP",
                "PDMROF_B",
                "DEP_MAX",
                "MAX_DEP_AREA_FRAC",
            ),
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(
                        P.X17,
                        P.X18,
                        P.X13,
                        P.X12,
                        0,
                        P.X01,
                        0,
                        P.X09,
                        P.X07,
                        P.X14,
                        P.X08,
                    ),
                )
            ],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": (
                "SAI_HT_RATIO",
                "MAX_CAPACITY",
                "MAX_SNOW_CAPACITY",
                "TFRAIN",
                "TFSNOW",
            ),
            "pl": [PL(name="FOREST", values=(0, 10000, 10000, 1, 1))],
        },
        alias="VegetationParameterList",
    )
    sub_basin_properties: rc.SubBasinProperties = Field(
        {"parameters": ["TIME_CONC"], "records": [{"sb_id": 1, "values": (P.X15,)}]},
        alias="SubBasinProperties",
    )

    hru_state_variable_table: rc.HRUStateVariableTable = Field(
        [
            rc.HRUState(
                hru_id=1,
                data={
                    "SOIL[0]": 10,
                    "SOIL[1]": 10,
                    "DEPRESSION": 20,
                },
            ),
        ],
        alias="HRUStateVariableTable",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)

    def __init__(self, **data):
        super().__init__(**data)

        if self.gauge:
            for gauge in self.gauge:
                gauge.snow_correction = self.params.X21



import numpy as np
import pandas as pd
import pytest

from ravenpy.utilities.regionalization import (
    distance,
    read_gauged_params,
    read_gauged_properties,
    regionalize,
)


class TestRegionalization:
    def test_full_example(self, symbolic_config):
        name, config = symbolic_config
        method = "SP_IDW"

        if name not in ["GR4JCN", "HMETS", "Mohyse"]:
            pytest.skip(f"Model {name} is not supported.")

        nash, params = read_gauged_params(name.upper())
        variables = ["latitude", "longitude", "area", "forest"]
        props = read_gauged_properties(variables)
        ungauged_props = {
            "latitude": 40.4848,
            "longitude": -103.3659,
            "area": 4250.6,
            "forest": 0.4,
        }

        qsim, ens = regionalize(
            config=config,
            method=method,
            nash=nash,
            params=params,
            props=props,
            target_props=ungauged_props,
            size=2,
            min_NSE=0.6,
        )

        assert qsim.max() > 1
        assert len(ens) == 2
        assert "members" in ens.dims
        assert "param" in ens.dims


class TestGreatCircle:
    def test_haversine_vector(self):
        places = dict()
        places["Ouagadougou"] = dict(latitude=12.366667, longitude=-1.533333)
        places["Nairobi"] = dict(latitude=-1.286389, longitude=36.817222)
        places["Cairo"] = dict(latitude=30.033333, longitude=31.233333)
        df = pd.DataFrame.from_dict(places).T
        algiers = pd.Series(dict(latitude=36.753889, longitude=3.058889))

        great_circle_vectors = distance(df, algiers)

        np.testing.assert_array_almost_equal(great_circle_vectors, [2750.249, 5478.388, 2709.166], 3)



import datetime as dt

import pytest
import xarray as xr

from ravenpy.config import commands as rc
from ravenpy.config.emulators import (
    GR4JCN,
    HBVEC,
    HMETS,
    HYPR,
    SACSMA,
    Blended,
    CanadianShield,
    Mohyse,
)


# Alternative names for variables in meteo forcing file
alt_names = {
    "RAINFALL": "rain",
    "TEMP_MIN": "tmin",
    "TEMP_MAX": "tmax",
    "PET": "pet",
    "HYDROGRAPH": "qobs",
    "SNOWFALL": "snow",
}

# Names of emulators to be included in tests
names = [
    "GR4JCN",
    "HMETS",
    "Mohyse",
    "HBVEC",
    "CanadianShield",
    "HYPR",
    "SACSMA",
    "Blended",
]

# Mapping of emulator names to configuration classes
configs = {
    "GR4JCN": GR4JCN,
    "HMETS": HMETS,
    "Mohyse": Mohyse,
    "HBVEC": HBVEC,
    "CanadianShield": CanadianShield,
    "HYPR": HYPR,
    "SACSMA": SACSMA,
    "Blended": Blended,
}

# Model parameters that match NSE criteria computed by Juliane
params = {
    "GR4JCN": [0.529, -3.396, 407.29, 1.072, 16.9, 0.947],
    "HMETS": [
        9.5019,
        0.2774,
        6.3942,
        0.6884,
        1.2875,
        5.4134,
        2.3641,
        0.0973,
        0.0464,
        0.1998,
        0.0222,
        -1.0919,
        2.6851,
        0.3740,
        1.0000,
        0.4739,
        0.0114,
        0.0243,
        0.0069,
        310.7211,
        916.1947,
    ],
    "Mohyse": (
        1.0,
        0.0468,
        4.2952,
        2.658,
        0.4038,
        0.0621,
        0.0273,
        0.0453,
        0.9039,
        5.6167,
    ),
    "HBVEC": (
        0.05984519,
        4.072232,
        2.001574,
        0.03473693,
        0.09985144,
        0.506052,
        3.438486,
        38.32455,
        0.4606565,
        0.06303738,
        2.277781,
        4.873686,
        0.5718813,
        0.04505643,
        0.877607,
        18.94145,
        2.036937,
        0.4452843,
        0.6771759,
        1.141608,
        1.024278,
    ),
    "CanadianShield": (
        4.72304300e-01,
        8.16392200e-01,
        9.86197600e-02,
        3.92699900e-03,
        4.69073600e-02,
        4.95528400e-01,
        6.803492000e00,
        4.33050200e-03,
        1.01425900e-05,
        1.823470000e00,
        5.12215400e-01,
        9.017555000e00,
        3.077103000e01,
        5.094095000e01,
        1.69422700e-01,
        8.23412200e-02,
        2.34595300e-01,
        7.30904000e-02,
        1.284052000e00,
        3.653415000e00,
        2.306515000e01,
        2.402183000e00,
        2.522095000e00,
        5.80344900e-01,
        1.614157000e00,
        6.031781000e00,
        3.11129800e-01,
        6.71695100e-02,
        5.83759500e-05,
        9.824723000e00,
        9.00747600e-01,
        8.04057300e-01,
        1.179003000e00,
        7.98001300e-01,
    ),
    "HYPR": (
        -1.856410e-01,
        2.92301100e00,
        3.1194200e-02,
        4.3982810e-01,
        4.6509760e-01,
        1.1770040e-01,
        1.31236800e01,
        4.0417950e-01,
        1.21225800e00,
        5.91273900e01,
        1.6612030e-01,
        4.10501500e00,
        8.2296110e-01,
        4.15635200e01,
        5.85111700e00,
        6.9090140e-01,
        9.2459950e-01,
        1.64358800e00,
        1.59920500e00,
        2.51938100e00,
        1.14820100e00,
    ),
    "SACSMA": (
        0.0100000,
        0.0500000,
        0.3000000,
        0.0500000,
        0.0500000,
        0.1300000,
        0.0250000,
        0.0600000,
        0.0600000,
        1.0000000,
        40.000000,
        0.0000000,
        0.0000000,
        0.1000000,
        0.0000000,
        0.0100000,
        1.5000000,
        0.4827523,
        4.0998200,
        1.0000000,
        1.0000000,
    ),
    "Blended": (
        2.930702e-02,
        2.211166e00,
        2.166229e00,
        0.0002254976,
        2.173976e01,
        1.565091e00,
        6.211146e00,
        9.313578e-01,
        3.486263e-02,
        0.251835,
        0.0002279250,
        1.214339e00,
        4.736668e-02,
        0.2070342,
        7.806324e-02,
        -1.336429e00,
        2.189741e-01,
        3.845617e00,
        2.950022e-01,
        4.827523e-01,
        4.099820e00,
        1.283144e01,
        5.937894e-01,
        1.651588e00,
        1.705806,
        3.719308e-01,
        7.121015e-02,
        1.906440e-02,
        4.080660e-01,
        9.415693e-01,
        -1.856108e00,
        2.356995e00,
        1.0e00,
        1.0e00,
        7.510967e-03,
        5.321608e-01,
        2.891977e-02,
        9.605330e-01,
        6.128669e-01,
        9.558293e-01,
        1.008196e-01,
        9.275730e-02,
        7.469583e-01,
    ),
}


@pytest.fixture(scope="session")
def gr4jcn_config(yangtze, salmon_hru) -> (GR4JCN, params):
    """Return symbolic config and params for basic gr4jcn."""
    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    yield (
        GR4JCN(
            Gauge=[
                rc.Gauge.from_nc(
                    salmon_file,
                    data_type=["RAINFALL", "TEMP_MIN", "TEMP_MAX", "SNOWFALL"],
                    alt_names=alt_names,
                    data_kwds={"ALL": {"elevation": salmon_hru["land"]["elevation"]}},
                ),
            ],
            ObservationData=[rc.ObservationData.from_nc(salmon_file, alt_names="qobs")],
            HRUs=[salmon_hru["land"]],
            StartDate=dt.datetime(2000, 1, 1),
            Duration=15,
            EvaluationMetrics=("NASH_SUTCLIFFE",),
        ),
        params["GR4JCN"],
    )


@pytest.fixture(scope="session", params=names)
def symbolic_config(yangtze, salmon_hru, request):
    """Emulator configuration instantiated with symbolic parameters."""
    name = request.param
    cls = configs[name]
    data_type = ["RAINFALL", "TEMP_MIN", "TEMP_MAX", "SNOWFALL"]

    # Extra attributes for gauges
    gextras = {"ALL": {"elevation": salmon_hru["land"]["elevation"]}}

    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    if name in ["HBVEC", "HYPR"]:
        with xr.open_dataset(salmon_file) as ds:
            gextras["ALL"]["monthly_ave_temperature"] = ((ds.tmin + ds.tmax) / 2).groupby("time.month").mean().values.tolist()
            gextras["ALL"]["monthly_ave_evaporation"] = ds.pet.groupby("time.month").mean().values.tolist()

    # Extra attributes for emulator
    if name in ["CanadianShield", "HYPR", "SACSMA", "Blended"]:
        salmon_hru["slope"] = 0.01234

    hrus = [salmon_hru["land"]]

    extras = {}
    if name in ["GR4JCN"]:
        extras.update({"GlobalParameter": {"AVG_ANNUAL_RUNOFF": 208.480}})
    if name in ["HMETS"]:
        data_type.extend(["PET"])

    if name in ["CanadianShield"]:
        hrus = [salmon_hru["land"], salmon_hru["land"]]
        extras["SuppressOutput"] = True

    yield (
        name,
        cls(
            Gauge=[
                rc.Gauge.from_nc(
                    salmon_file,
                    data_type=data_type,
                    alt_names=alt_names,
                    data_kwds=gextras,
                ),
            ],
            ObservationData=[rc.ObservationData.from_nc(salmon_file, alt_names="qobs")],
            HRUs=hrus,
            StartDate=dt.datetime(2000, 1, 1),
            EndDate=dt.datetime(2002, 1, 1),
            RunName="test",
            EvaluationMetrics=("NASH_SUTCLIFFE",),
            **extras,
        ),
    )


@pytest.fixture(scope="session")
def numeric_config(symbolic_config):
    """Emulator configuration instantiated with numeric parameters."""
    name, conf = symbolic_config
    yield name, conf.set_params(params[name])


@pytest.fixture(scope="session")
def minimal_emulator(yangtze, salmon_hru):
    """Return the config for a single emulator."""
    cls = configs["HMETS"]
    data_type = ["RAINFALL", "TEMP_MIN", "TEMP_MAX", "SNOWFALL", "PET"]

    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    yield cls(
        params=params["HMETS"],
        Gauge=[
            rc.Gauge.from_nc(
                salmon_file,
                data_type=data_type,
                alt_names=alt_names,
            ),
        ],
        ObservationData=[rc.ObservationData.from_nc(salmon_file, alt_names="qobs")],
        HRUs=[salmon_hru["land"]],
        StartDate=dt.datetime(2000, 1, 1),
        Duration=15,
    )


@pytest.fixture(scope="session")
def config_rv(tmp_path_factory, numeric_config):
    """Directory with emulator configuration files written to disk."""
    name, conf = numeric_config
    out = tmp_path_factory.mktemp(name) / "config"
    conf.write_rv(out)
    yield name, out



import pytest
import xarray as xr
from xclim import set_options


class TestNBGraphs:
    hydrographs = "hydro_simulations/raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc"

    nbg = pytest.importorskip("ravenpy.utilities.nb_graphs")

    def test_hydrograph(self, yangtze):
        with xr.open_dataset(yangtze.fetch(self.hydrographs)) as ds:
            self.nbg.hydrographs(ds)

    def test_mean_annual_hydrograph(self, yangtze):
        with xr.open_dataset(yangtze.fetch(self.hydrographs)) as ds:
            self.nbg.mean_annual_hydrograph(ds)

    def test_spaghetti_annual_hydrograph(self, yangtze):
        with xr.open_dataset(yangtze.fetch(self.hydrographs)) as ds:
            self.nbg.spaghetti_annual_hydrograph(ds)

    def test_ts_fit_graph(self, yangtze):
        from xclim.indicators.generic import fit, stats

        with xr.open_dataset(yangtze.fetch(self.hydrographs)) as ds:
            ts = stats(ds.q_sim.load(), op="max", freq="ME")
        with set_options(check_missing="skip"):
            params = fit(ts, dist="gamma")
        self.nbg.ts_fit_graph(ts, params)



from collections.abc import Sequence
from typing import Literal, Union

from pydantic import Field, field_validator

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.commands import HRU, Process, SoilModel
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


class LandHRU(HRU):
    land_use_class: str = "Landuse_Land_HRU"
    veg_class: str = "Veg_Land_HRU"
    soil_profile: str = "Soil_Land_HRU"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["land"] = "land"


class LakeHRU(HRU):
    land_use_class: str = "Landuse_Lake_HRU"
    veg_class: str = "Veg_Lake_HRU"
    soil_profile: str = "Soil_Lake_HRU"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["lake"] = "lake"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[Union[LandHRU, LakeHRU]]


class BasicRoute(Config):
    """Raven configuration performing routing only."""

    hrus: HRUs = Field([LandHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "BasicRoute"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    catchment_route: o.CatchmentRoute = Field(
        "ROUTE_DUMP",
        alias="CatchmentRoute",
        description="Catchment routing method, used to convey water from the catchment tributaries and rivulets to the subbasin outlets.",
    )
    routing: o.Routing = Field(
        "ROUTE_DIFFUSIVE_WAVE",
        alias="Routing",
        description="Channel routing method which is used to transport water from upstream to downstream within the main subbasin channels.",
    )
    precip_icept_frac: o.PrecipIceptFract = Field(
        "PRECIP_ICEPT_NONE",
        alias="PrecipIcepFract",
        description="Estimation of the precipitation interception fraction. In this routing "
        "model, stream input(s) are pretending to be precipitation going into "
        "Raven.",
    )
    potential_melt_method: o.PotentialMeltMethod = Field(
        "POTMELT_NONE",
        alias="PotentialMeltMethod",
        description="Estimation of the potential snow melt. In this routing model, snow melt processes are not "
        "relevant, thus using DEFAULT POTMELT_NONE method.",
    )
    soil_model: SoilModel = Field(1, alias="SoilModel", description="Single soil layer structure")
    hydrologic_processes: Sequence[Process] = Field(
        [
            p.Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="PONDED_WATER"),
            p.Flush(source="PONDED_WATER", to="SURFACE_WATER"),
        ],
        alias="HydrologicProcesses",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)



from collections import defaultdict
from pathlib import Path

import click

from ravenpy.utilities import gis_import_error_message


try:
    import geopandas as gpd
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e


@click.command()
@click.argument("input-file", type=click.Path(exists=True))
@click.argument("gauge-id", type=str)
@click.option("-o", "--output", type=click.Path(), help="Name of the output shapefile.")
def collect_subbasins_upstream_of_gauge(
    input_file: str,
    gauge_id: str,
    output: str,
):
    """
    Find the subbasins upstream of a gauge from a Routing Product shapefile, and save them in a new shapefile.

    Parameters
    ----------
    input_file : str
        Routing Product shapefile (e.g. "drainage_region_0003_v2-1/finalcat_info_v2-1.shp").
    gauge_id : str
        ID of the target gauge, to be found in the "Obs_NM" column (e.g. "02LE024").
    output : str
        Name of the output shapefile.
    """
    if Path(input_file).suffix == ".zip":
        input_file = f"zip://{input_file}"
    gdf = gpd.read_file(input_file)

    if gauge_id not in gdf.Obs_NM.values:
        raise click.ClickException(f"Cannot find gauge {gauge_id} in `Obs_NM` column")

    if not output:
        p = Path(input_file)
        output_file = p.parent / f"{p.stem}_upstream_of_gauge_{gauge_id}.shp"
    else:
        output_file = output

    downsubid_to_subids = defaultdict(set)

    for _, r in gdf.iterrows():
        downsubid_to_subids[r.DowSubId].add(r.SubId)

    # Starting from the SubId of the gauge, iteratively expand the set of upstream subbasin SubIds
    upstream_subids = {gdf[gdf.Obs_NM == gauge_id].iloc[0].SubId}
    prev = upstream_subids.copy()
    while True:
        curr = set()
        for did in prev:
            curr |= downsubid_to_subids[did]
        if curr:
            upstream_subids |= curr
            prev = curr
        else:
            break

    gdf_upstream = gdf[gdf.SubId.isin(upstream_subids)]

    gpd.GeoDataFrame(gdf_upstream).to_file(output_file)

    click.echo(f"Found {len(gdf_upstream)} upstream subbasins, saved them in {output_file}")



import pytest
import xsdba
from xarray.coding.calendar_ops import convert_calendar


# FIXME: This doesn't test Raven functionality; Should we move it to xclim?
@pytest.mark.skip(reason="This test is not testing Raven functionality")
class TestBiasCorrect:
    def test_bias_correction(self, open_dataset):
        ds_fut_sub = open_dataset("cmip5/nasa_nex-gddp-1.0_day_inmcm4_historical+rcp85_nex-gddp_2070-2071_subset.nc")

        ds_ref_sub = open_dataset("cmip5/nasa_nex-gddp-1.0_day_inmcm4_historical+rcp45_nex-gddp_1971-1972_subset.nc")
        ds_ref_sub = convert_calendar(ds_ref_sub, "noleap")

        ds_his_sub = open_dataset("nrcan/NRCAN_1971-1972_subset.nc")

        ds_his_sub = convert_calendar(ds_his_sub, "noleap")
        group = xsdba.Grouper("time.month")
        # Train the model to find the correction factors
        Adj = xsdba.DetrendedQuantileMapping.train(
            ref=ds_ref_sub["pr"],
            hist=ds_his_sub["pr"],
            nquantiles=50,
            kind="+",
            group=group,
        )

        # Apply the factors to the future data to bias-correct
        Adj.adjust(ds_fut_sub["pr"], interp="linear")

        # Repeat for temperature max
        Adj = xsdba.DetrendedQuantileMapping.train(
            ref=ds_ref_sub["tasmax"],
            hist=ds_his_sub["tasmax"],
            nquantiles=50,
            kind="+",
            group=group,
        )

        # Apply the factors to the future data to bias-correct
        Adj.adjust(ds_fut_sub["tasmax"], interp="linear")

        # Repeat for tasmin
        Adj = xsdba.DetrendedQuantileMapping.train(
            ref=ds_ref_sub["tasmin"],
            hist=ds_his_sub["tasmin"],
            nquantiles=50,
            kind="+",
            group=group,
        )

        Adj.adjust(ds_fut_sub["tasmin"], interp="linear")

        # TODO: Add numerical check



from collections.abc import Sequence
from typing import Literal, Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Params, Sym, SymConfig
from ravenpy.config.commands import (
    HRU,
    PL,
    LandUseClasses,
    LandUseParameterList,
    Process,
    RainSnowTransition,
    SoilClasses,
    SoilModel,
    SoilParameterList,
    SoilProfiles,
    VegetationClasses,
)
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


@dataclass(config=SymConfig)
class P(Params):
    GR4J_X1: Sym = Variable("GR4J_X1")
    GR4J_X2: Sym = Variable("GR4J_X2")
    GR4J_X3: Sym = Variable("GR4J_X3")
    GR4J_X4: Sym = Variable("GR4J_X4")
    CEMANEIGE_X1: Sym = Variable("CEMANEIGE_X1")
    CEMANEIGE_X2: Sym = Variable("CEMANEIGE_X2")


# Note that the `hru_type` field is not part of the Raven configuration specification, but used here to automatically
# detect whether an HRU is land or lake. Just pass `hru_type` within the dictionary of attributes used to instantiate
# HRUs.


class LandHRU(HRU):
    land_use_class: str = "LU_ALL"
    veg_class: str = "VEG_ALL"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["land"] = "land"


class LakeHRU(HRU):
    land_use_class: str = "LU_WATER"
    veg_class: str = "VEG_WATER"
    soil_profile: str = "LAKE"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["lake"] = "lake"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[Union[LandHRU, LakeHRU]]


class GR4JCN(Config):
    """
    GR4J + Cemaneige global hydrological model.

    References
    ----------
    Perrin, C., C. Michel and V. Andréassian (2003). Improvement of a parsimonious model for streamflow simulation.
    Journal of Hydrology, 279(1-4), 275-289. doi: 10.1016/S0022-1694(03)00225-7.

    Valéry, Audrey, Vazken Andréassian, and Charles Perrin. 2014. “’As Simple as Possible but Not Simpler’: What Is
    Useful in a Temperature-Based Snow-Accounting Routine? Part 2 - Sensitivity Analysis of the Cemaneige Snow
    Accounting Routine on 380 Catchments.” Journal of Hydrology, no. 517(0): 1176–87,
    doi: 10.1016/j.jhydrol.2014.04.058.
    """

    params: P = P()
    hrus: HRUs = Field([LandHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "GR4JCN"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    uniform_initial_conditions: Union[dict[str, Sym], None] = Field(
        {"SOIL[0]": P.GR4J_X1 * 1000 / 2, "SOIL[1]": 15},
        alias="UniformInitialConditions",
    )
    evaporation: o.Evaporation = Field("PET_OUDIN", alias="Evaporation")
    rain_snow_fraction: o.RainSnowFraction = Field("RAINSNOW_DINGMAN", alias="RainSnowFraction")
    soil_model: Union[int, SoilModel] = Field(4, alias="SoilModel")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_DUMP", alias="CatchmentRoute")
    potential_melt: o.PotentialMeltMethod = Field("POTMELT_DEGREE_DAY", alias="PotentialMeltMethod")
    oro_temp_correct: o.OroTempCorrect = Field(o.OroTempCorrect.SIMPLELAPSE, alias="OroTempCorrect")
    oro_precip_correct: o.OroPrecipCorrect = Field(o.OroPrecipCorrect.SIMPLELAPSE, alias="OroPrecipCorrect")

    hydrologic_processes: Sequence[Union[Process, p.Conditional]] = Field(
        [
            p.Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.SnowTempEvolve(algo="SNOTEMP_NEWTONS", source="SNOW_TEMP"),
            p.SnowBalance(algo="SNOBAL_CEMA_NIEGE", source="SNOW", to="PONDED_WATER"),
            p.OpenWaterEvaporation(algo="OPEN_WATER_EVAP", source="PONDED_WATER", to="ATMOSPHERE"),
            p.Infiltration(algo="INF_GR4J", source="PONDED_WATER", to="MULTIPLE"),
            p.SoilEvaporation(algo="SOILEVAP_GR4J", source="SOIL[0]", to="ATMOSPHERE"),
            p.Percolation(algo="PERC_GR4J", source="SOIL[0]", to="SOIL[2]"),
            p.Flush(
                algo="RAVEN_DEFAULT",
                source="SURFACE_WATER",
                to="SOIL[2]",
            ),
            p.Split(
                algo="RAVEN_DEFAULT",
                source="SOIL[2]",
                to=("CONVOLUTION[0]", "CONVOLUTION[1]"),
                p=0.9,
            ),
            p.Convolve(algo="CONVOL_GR4J_1", source="CONVOLUTION[0]", to="SOIL[1]"),
            p.Convolve(algo="CONVOL_GR4J_2", source="CONVOLUTION[1]", to="SOIL[2]"),
            p.Percolation(algo="PERC_GR4JEXCH", source="SOIL[1]", to="SOIL[3]"),
            p.Percolation(algo="PERC_GR4JEXCH2", source="SOIL[2]", to="SOIL[3]"),
            p.Flush(algo="RAVEN_DEFAULT", source="SOIL[2]", to="SURFACE_WATER"),
            p.Baseflow(algo="BASE_GR4J", source="SOIL[1]", to="SURFACE_WATER"),
        ],
        alias="HydrologicProcesses",
    )

    rain_snow_transition: RainSnowTransition = Field({"temp": 0, "delta": 1}, alias="RainSnowTransition")
    global_parameter: dict[str, Sym] = Field(
        {
            "PRECIP_LAPSE": 0.0004,
            "ADIABATIC_LAPSE": 0.0065,
            "AVG_ANNUAL_SNOW": P.CEMANEIGE_X1,
            "AIRSNOW_COEFF": 1 - P.CEMANEIGE_X2,
        },
        alias="GlobalParameter",
    )

    soil_classes: SoilClasses = Field(
        [
            {"name": "SOIL_PROD"},
            {"name": "SOIL_ROUT"},
            {"name": "SOIL_TEMP"},
            {"name": "SOIL_GW"},
            {"name": "AQUIFER"},
        ],
        alias="SoilClasses",
    )

    soil_parameter_list: SoilParameterList = Field(
        {
            "parameters": ("POROSITY", "GR4J_X3", "GR4J_X2"),
            "pl": [PL(name="[DEFAULT]", values=(1, P.GR4J_X3, P.GR4J_X2))],
        },
        alias="SoilParameterList",
    )

    soil_profiles: SoilProfiles = Field(
        [
            dict(
                name="DEFAULT_P",
                soil_classes=["SOIL_PROD", "SOIL_ROUT", "SOIL_TEMP", "SOIL_GW"],
                thicknesses=[P.GR4J_X1, 0.3, 1, 1],
            ),
            dict(name="LAKE"),
        ],
        alias="SoilProfiles",
    )

    vegetation_classes: VegetationClasses = Field(
        [{"name": "VEG_ALL"}, {"name": "VEG_WATER"}],
        alias="VegetationClasses",
    )

    land_use_classes: LandUseClasses = Field(
        [{"name": "LU_ALL"}, {"name": "LU_WATER"}],
        alias="LandUseClasses",
    )

    land_use_parameter_list: LandUseParameterList = Field(
        {
            "parameters": ("GR4J_X4", "MELT_FACTOR"),
            "pl": [PL(name="[DEFAULT]", values=(P.GR4J_X4, 7.73))],
        },
        alias="LandUseParameterList",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)



import datetime as dt
import pathlib
import sys

import pytest
import xarray as xr

from ravenpy import Emulator, EnsembleReader
from ravenpy.config import commands as rc
from ravenpy.config.emulators import GR4JCN
from ravenpy.utilities.forecasting import (
    hindcast_climatology_esp,
    to_climpred_hindcast_ensemble,
    warm_up,
)


"""
Test to perform a hindcast using Caspar data on THREDDS.
Currently only runs GEPS, eventually will run GEPS, GDPS, REPS and RDPS.
To do so will need to add the actual data from Caspar, currently being downloaded
but this is a good proof of concept.
"""


class TestHindcasting:
    def test_hindcasting_GEPS(self, salmon_hru, tmp_path, yangtze):
        ts20 = yangtze.fetch("caspar_eccc_hindcasts/geps_watershed.nc")

        hru = salmon_hru["land"]
        data_kwds = {
            "ALL": {"Latitude": hru["latitude"], "Longitude": hru["longitude"]},
            "PRECIP": {
                "Deaccumulate": True,
                "TimeShift": -0.25,
                "LinearTransform": {"scale": 1000},
            },
        }

        pre_conf = GR4JCN(
            StartDate=dt.datetime(2018, 6, 10),
            EndDate=dt.datetime(2018, 6, 14),
            HRUs=[
                hru,
            ],
            params=(0.529, -3.396, 407.29, 1.072, 16.9, 0.947),
            GlobalParameter={"AVG_ANNUAL_RUNOFF": 208.480},
            Gauge=[
                rc.Gauge.from_nc(ts20, data_type=["PRECIP", "TEMP_AVE"], data_kwds=data_kwds),
            ],
        )

        init_conf = warm_up(pre_conf, duration=8, workdir=tmp_path / "wup")

        # Create gauges for each member and run the model
        nm = 3
        ens = []
        for i in range(nm):
            g = rc.Gauge.from_nc(
                ts20,
                station_idx=i + 1,
                data_type=["PRECIP", "TEMP_AVE"],
                data_kwds=data_kwds,
            )
            assert g.data[0].read_from_netcdf.station_idx == i + 1
            for d in g.data:
                if d.data_type == "PRECIP":
                    assert d.read_from_netcdf.linear_transform.scale == 1000

            mem_conf = init_conf.model_copy(
                update=dict(
                    gauge=[g],
                ),
            )

            # This segfaults
            e = Emulator(
                config=mem_conf,
                workdir=tmp_path / f"m{i:02}",
            )
            ens.append(e.run())

        out = EnsembleReader(runs=ens)

        # The model now has the forecast data generated, and it has 5 days of forecasts.
        assert len(out.hydrograph.member) == 3
        assert len(out.hydrograph.time) == 5

    # Skip if using Python3.10
    @pytest.mark.skipif(
        (3, 11) > sys.version_info >= (3, 10),
        reason="climpred is unstable in Python 3.10",
    )
    def test_climpred_hindcast_verif(self, salmon_hru, tmp_path, yangtze):
        ts = pathlib.Path(yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"))
        # Make a local copy to evade double-ownership of file - first file
        ts_tmp1 = tmp_path / "salmon_river_near_prince_george-tmp1.nc"
        ts_tmp1.write_bytes(ts.read_bytes())
        # Make a local copy to evade double-ownership of file - second file
        ts_tmp2 = tmp_path / "salmon_river_near_prince_george-tmp2.nc"
        ts_tmp2.write_bytes(ts.read_bytes())

        # This is the forecast start date, on which the forecasts will be launched.
        start_date = dt.datetime(1980, 6, 1)

        # Provide the length of the forecast, in days:
        forecast_duration = 100

        # Define HRU to build the hydrological model
        hru = salmon_hru["land"]

        # Set alternative names for netCDF variables
        alt_names = {
            "TEMP_MIN": "tmin",
            "TEMP_MAX": "tmax",
            "RAINFALL": "rain",
            "SNOWFALL": "snow",
        }

        # Data types to extract from netCDF
        data_type = ["TEMP_MAX", "TEMP_MIN", "RAINFALL", "SNOWFALL"]
        data_kwds = {
            "ALL": {
                "elevation": hru["elevation"],  # No need for lat/lon as they are included in the netcdf file already
            }
        }

        # Model configuration
        model_config = GR4JCN(
            params=[0.529, -3.396, 407.29, 1.072, 16.9, 0.947],
            Gauge=[
                rc.Gauge.from_nc(
                    ts_tmp1,
                    data_type=data_type,
                    alt_names=alt_names,
                    data_kwds=data_kwds,
                )
            ],
            HRUs=[hru],
            StartDate=start_date,
            Duration=forecast_duration,
            GlobalParameter={"AVG_ANNUAL_RUNOFF": 208.480},
        )

        hindcasts = hindcast_climatology_esp(
            config=model_config,  # Note that the forecast duration is already set-up in the model_config above.
            warm_up_duration=365,  # number of days for the warm-up
            years=[1985, 1986, 1987, 1988, 1989, 1990],
            hindcast_years=[2001, 2002, 2003, 2004, 2005, 2006, 2007],
        )

        q_obs = xr.open_dataset(ts_tmp2)

        # However, our simulated streamflow is named "q_sim" and climpred requires the observation to be named the same thing
        # so let's rename it. While we're at it, we need to make sure that the identifier is the same. In our observation
        # dataset, it is called "nstations" but in our simulated streamflow it's called "nbasins". Here we standardize.
        q_obs = q_obs.rename({"qobs": "q_sim", "nstations": "nbasins"})

        # Make the hindcasting object we can use to compute statistics and metrics
        hindcast_object = to_climpred_hindcast_ensemble(hindcasts, q_obs)

        # This function is used to convert to binary to see if yes/no forecast is larger than observations
        def pos(x):
            return x > 0  # Check for binary outcome

        # Rank histogram verification metric
        rank_histo_verif = hindcast_object.verify(
            metric="rank_histogram",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
        )
        assert "q_sim" in rank_histo_verif
        assert rank_histo_verif.q_sim.shape[0] - 1 == forecast_duration

        # CRPS verification metric
        crps_verif = hindcast_object.verify(
            metric="crps",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
        )
        assert "q_sim" in crps_verif
        assert crps_verif.q_sim.shape[0] - 1 == forecast_duration

        # TODO: Fix this part
        """
        reliability_verif = hindcast_object.verify(
            metric="reliability",
            comparison="m2o",
            dim=["member", "init"],
            alignment="same_inits",
            logical=pos,
        )

        assert "flow" in reliability_verif
        assert reliability_verif.flow.shape[0] == forecast_duration
        """



"""Main module."""

import collections
import os
import subprocess  # noqa: S404
import tempfile
import warnings
from pathlib import Path
from typing import Optional, Union
from warnings import warn

import xarray as xr

from ravenpy import RAVEN_EXEC_PATH

from .config import parsers
from .config.rvs import Config


class Emulator:
    def __init__(
        self,
        config: Config,
        workdir: Optional[Union[str, os.PathLike]] = None,
        modelname: Optional[str] = None,
        overwrite: bool = False,
    ):
        """
        Convenience class to work with the Raven modeling framework.

        Parameters
        ----------
        config : Config
            Emulator Config instance fully parameterized, i.e. without symbolic expressions.
        workdir : Path or str
            Path to rv files and model outputs. If None, create a temporary directory.
        modelname : str, None
            File name stem of configuration files: `<modelname>.rv*`.
        overwrite : bool
            If True, overwrite existing files.
        """
        self._config = config.model_copy(deep=True)
        self._workdir = Path(workdir or tempfile.mkdtemp())
        self._modelname = modelname
        self.overwrite = overwrite
        self._output = None  # Model output path
        self._output_path = None

        # Write model config files
        self._rv = self._config.write_rv(workdir=self.workdir, modelname=self.modelname, overwrite=overwrite)
        # Grab modelname in case it was set by config.write_rv
        self._modelname = self._rv["rvi"].stem

    def run(self, overwrite: bool = False) -> "OutputReader":
        """
        Run the model.

        This will write RV files if not already done.

        Parameters
        ----------
        overwrite : bool
            If True, overwrite existing files.
        """
        self._output_path = run(self.modelname, self.workdir, "output", overwrite=overwrite)
        self._output = OutputReader(self.config.run_name, path=self._output_path)
        return self._output

    @property
    def config(self) -> Config:
        """Read-only model configuration."""
        return self._config

    @property
    def workdir(self) -> Path:
        """Path to RV files and output subdirectory."""
        return self._workdir

    @property
    def output_path(self) -> Optional[Path]:
        """Path to model outputs."""
        if self._output_path is not None:
            return self._output_path
        warnings.warn("`output_path` not set. Model must be run first.", stacklevel=2)
        return None

    @property
    def modelname(self) -> str:
        """File name stem of configuration files."""
        return self._modelname

    @property
    def output(self) -> "OutputReader":
        """Return simulation output object."""
        return self._output

    def resume(self, timestamp: bool = True) -> Config:
        """
        Return new model configuration using state variables from the end of the run.

        Parameters
        ----------
        timestamp : bool
            If False, ignore time stamp information in the solution.
            If True, the solution will set StartDate to the solution's timestamp.
        """
        return self.config.set_solution(self.output.files["solution"], timestamp=timestamp)


class OutputReader:
    def __init__(self, run_name: Optional[str] = None, path: Optional[Union[str, Path]] = None):
        """
        Class facilitating access to Raven model output.

        Parameters
        ----------
        run_name : str, optional
            Simulation name, if any is specified by the `RunName` configuration.
        path : str or Path
            Output directory where model results are stored. Defaults to the current directory.
        """
        self._run_name = run_name
        self._path = Path(path) if path else Path.cwd()
        self._files = parsers.output_files(self._run_name, self._path)
        self._nc_hydrograph = None
        self._nc_storage = None

        # TODO: Check if no files are found. Otherwise we get cryptic errors.
        # if self._files["hydrograph"]

    @property
    def files(self) -> dict:
        """Return paths to output files."""
        return self._files

    @property
    def solution(self) -> Optional[dict]:
        """Return solution file content."""
        solution = self.files.get("solution")
        if solution:
            return parsers.parse_solution(solution)
        return None

    @property
    def diagnostics(self) -> Optional[dict]:
        """Return model diagnostics."""
        diag = self.files.get("diagnostics")
        if diag:
            return parsers.parse_diagnostics(diag)
        return None

    @property
    def hydrograph(self) -> xr.Dataset:
        """Return the hydrograph."""
        if self._nc_hydrograph is None:
            h = self.files.get("hydrograph")
            if h:
                self._nc_hydrograph = parsers.parse_nc(h)

        return self._nc_hydrograph

    @property
    def storage(self) -> xr.Dataset:
        """Return the storage variables."""
        if self._nc_storage is None:
            s = self.files.get("storage")
            if s:
                self._nc_storage = parsers.parse_nc(s)

        return self._nc_storage

    @property
    def messages(self) -> Optional[str]:
        msg = self.files.get("messages")
        if msg:
            return msg.read_text()
        return None

    @property
    def path(self) -> Path:
        """Path to output directory."""
        return self._path


class EnsembleReader:
    def __init__(
        self,
        *,
        run_name: Optional[str] = None,
        paths: Optional[list[Union[str, os.PathLike]]] = None,
        runs: Optional[list[OutputReader]] = None,
        dim: str = "member",
    ):
        """
        Class facilitating access to ensemble of Raven outputs.

        Parameters
        ----------
        run_name : str, None
            Name given to simulation, if any.
        paths : list[str or Path], optional
            List of output paths. Defaults to all directories in current directory.
        runs : List[OutputReader], optional
            List of OutputReader instances.
        dim : str
            Name of concatenation dimension.
        """
        self._dim = dim
        if runs is not None:
            self._outputs = runs
            self._paths = [o.path for o in self._outputs]

        else:
            if paths is None:
                paths = [p for p in Path.cwd().iterdir() if p.is_dir()]
            self._paths = [Path(p) for p in paths]
            self._outputs = [OutputReader(run_name, p) for p in self._paths]

    @property
    def files(self):
        out = collections.defaultdict(list)
        for o in self._outputs:
            for k, v in o.files.items():
                out[k].append(v)
        return out

    @property
    def storage(self):
        if len(self.files["storage"]) == 0:
            raise ValueError("No file found, make sure you have the right `run_name` and output `paths`.")

        return xr.concat([xr.open_dataset(f) for f in self.files["storage"]], dim=self._dim)

    @property
    def hydrograph(self):
        if len(self.files["hydrograph"]) == 0:
            raise ValueError("No file found, make sure you have the right `run_name` and output `paths`.")
        return xr.concat(
            [xr.open_dataset(f) for f in self.files["hydrograph"]],
            dim=self._dim,
            coords="different",
        )


def run(
    modelname: str,
    configdir: Union[str, Path],
    outputdir: Optional[Union[str, Path]] = None,
    overwrite: bool = True,
    verbose: bool = False,
) -> Path:
    """
    Run Raven given the path to an existing model configuration.

    Parameters
    ----------
    modelname : str
        Configuration files stem, i.e. the file name without extension.
    configdir : Path or str
        Path to configuration files directory.
    outputdir : Path or str, optional
        Path to model simulation output.
        If None, will write to configdir/output.
    overwrite : bool
        If True, overwrite existing files.
    verbose : bool
        If True, always display Raven warnings.
        If False, warnings will only be printed if an error occurs.

    Returns
    -------
    Path
        The path to the model outputs.
    """
    # Confirm configdir exists
    configdir = Path(configdir).absolute()
    if not configdir.exists():
        raise OSError("Workdir should include configuration files.")

    # Create outputdir
    outputdir = Path(outputdir or "output")
    if not outputdir.is_absolute():
        outputdir = (configdir / outputdir).absolute()

    if not outputdir.exists():
        Path(str(outputdir)).mkdir(parents=True)

    # Parse RunName
    rvi = (configdir / f"{modelname}.rvi").read_text()
    run_name = parsers.parse_rv(rvi, "RunName") or ""

    # Existing output files with the same :RunName - they would be overwritten
    files = outputdir.glob(f"{run_name}*.*")

    # Remove existing output files if overwrite is True
    for f in files:
        if f.is_file():
            if overwrite:
                f.unlink()
            else:
                raise FileExistsError("Output files using this `modelname` already exist. Use `overwrite=True` to remove them.")

    # Launch executable, wait for completion.
    cmd = [RAVEN_EXEC_PATH, modelname, "-o", str(outputdir)]

    process = subprocess.Popen(  # noqa: S603
        cmd,
        cwd=configdir,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    stdout, stderr = process.communicate(input="\n")
    return_code = process.wait()

    # Deal with errors and warnings
    messages = parsers.parse_raven_messages(outputdir / "Raven_errors.txt")

    if messages["ERROR"] or verbose:
        for msg in messages["WARNING"] + messages["ADVISORY"]:
            warn(msg, category=RavenWarning, stacklevel=2)

    if messages["ERROR"]:
        raise RavenError("\n".join([f"Config directory: {configdir}"] + messages["ERROR"]))

    if return_code != 0:
        raise OSError(f"Raven Error (code: {return_code}): \n{stdout}\n{stderr}")

    return outputdir


class RavenError(Exception):
    """
    RavenError exception class.

    An error that is meant to be raised whenever a message of type "ERROR" is found
    in the Raven_errors.txt file resulting from a Raven (i.e. the C program) run.
    """

    pass


class RavenWarning(Warning):
    """
    RavenWarning warning class.

    A warning corresponding to a message of type "WARNING" in the Raven_errors.txt
    file resulting from a Raven (i.e. the C program) run.
    """

    pass



from collections.abc import Sequence
from typing import Literal, Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Params, Sym, SymConfig
from ravenpy.config.commands import HRU, PL, Process
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


@dataclass(config=SymConfig)
class P(Params):
    X01: Sym = Variable("X01")  #
    X02: Sym = Variable("X02")
    X03: Sym = Variable("X03")
    X04: Sym = Variable("X04")
    X05: Sym = Variable("X05")
    X06: Sym = Variable("X06")
    X07: Sym = Variable("X07")
    X08: Sym = Variable("X08")
    X09: Sym = Variable("X09")
    X10: Sym = Variable("X10")


class LandHRU(HRU):
    land_use_class: str = "LU_ALL"
    veg_class: str = "VEG_ALL"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["land"] = "land"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[LandHRU]


class Mohyse(Config):
    """
    Modèle Hydrologique Simplifié à l'Extrême (MOHYSE)

    References
    ----------
    Fortin, V.; Turcotte, R. Le modèle hydrologique MOHYSE. In Note de Cours Pour SCA7420,
    Université du Québec à Montréal: Montréal, QC, Canada, 2007; p. 14.

    Troin, M., Arsenault, R. and Brissette, F., 2015. Performance and uncertainty evaluation
    of snow models on snowmelt flow simulations over a Nordic catchment (Mistassibi, Canada).
    Hydrology, 2(4), pp.289-317.
    """

    params: P = P()
    hrus: HRUs = Field([LandHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "Mohyse"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    soil_classes: rc.SoilClasses = Field([{"name": "TOPSOIL"}, {"name": "GWSOIL"}], alias="SoilClasses")
    land_use_classes: rc.LandUseClasses = Field(
        [{"name": "LU_ALL", "impermeable_frac": 0, "forest_coverage": 1}],
        alias="LandUseClasses",
    )
    vegetation_classes: rc.VegetationClasses = Field([{"name": "VEG_ALL"}], alias="VegetationClasses")
    soil_profiles: rc.SoilProfiles = Field(
        [
            {"name": "LAKE"},
            {"name": "ROCK"},
            {
                "name": "DEFAULT_P",
                "soil_classes": ["TOPSOIL", "GWSOIL"],
                "thicknesses": [P.X05, 10.0],
            },
        ],
        alias="SoilProfiles",
    )
    global_parameter: dict = Field(
        {"RAINSNOW_TEMP": -2, "TOC_MULTIPLIER": 1, "MOHYSE_PET_COEFF": P.X01},
        alias="GlobalParameter",
    )
    soil_parameter_list: rc.SoilParameterList = Field(
        {
            "parameters": [
                "POROSITY",
                "PET_CORRECTION",
                "HBV_BETA",
                "BASEFLOW_COEFF",
                "PERC_COEFF",
            ],
            "pl": [
                PL(name="TOPSOIL", values=(1.0, 1.0, 1.0, P.X07, P.X06)),
                PL(name="GWSOIL", values=(1.0, 1.0, 1.0, P.X08, 0.0)),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: rc.LandUseParameterList = Field(
        {
            "parameters": [
                "MELT_FACTOR",
                "AET_COEFF",
                "FOREST_SPARSENESS",
                "DD_MELT_TEMP",
            ],
            "pl": [PL(name="[DEFAULT]", values=(P.X03, P.X02, 0, P.X04))],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": ["SAI_HT_RATIO", "RAIN_ICEPT_PCT", "SNOW_ICEPT_PCT"],
            "pl": [PL(name="[DEFAULT]", values=(0, 0, 0))],
        },
        alias="VegetationParameterList",
    )
    soil_model: rc.SoilModel = Field(2, alias="SoilModel")
    potential_melt_method: o.PotentialMeltMethod = Field("POTMELT_DEGREE_DAY", alias="PotentialMeltMethod")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_GAMMA_CONVOLUTION", alias="CatchmentRoute")
    evaporation: o.Evaporation = Field("PET_MOHYSE", alias="Evaporation")
    direct_evaporation: bool = Field(True, alias="DirectEvaporation")
    rain_snow_fraction: o.RainSnowFraction = Field("RAINSNOW_DATA", alias="RainSnowFraction")
    hydrologic_processes: Sequence[Process] = Field(
        [
            p.SoilEvaporation(algo="SOILEVAP_LINEAR", source="SOIL[0]", to="ATMOSPHERE"),
            p.SnowBalance(algo="SNOBAL_SIMPLE_MELT", source="SNOW", to="PONDED_WATER"),
            p.Precipitation(algo="RAVEN_DEFAULT", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.Infiltration(algo="INF_HBV", source="PONDED_WATER", to="MULTIPLE"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[0]", to="SURFACE_WATER"),
            p.Percolation(algo="PERC_LINEAR", source="SOIL[0]", to="SOIL[1]"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[1]", to="SURFACE_WATER"),
        ]
    )
    sub_basin_properties: rc.SubBasinProperties = Field(
        {
            "parameters": ["GAMMA_SCALE", "GAMMA_SHAPE"],
            "records": [
                {"sb_id": "1", "values": (1 / P.X10, P.X09)},
            ],
        },
        alias="SubBasinProperties",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)



#!/usr/bin/env python3
# Created on Fri Jul 17 09:11:58 2020
# @author: ets
import datetime as dt
import logging
import os
import tempfile
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import climpred
import xarray as xr
from climpred import HindcastEnsemble

from ravenpy import Emulator, EnsembleReader
from ravenpy.config import commands as rc
from ravenpy.config.rvs import Config


LOGGER = logging.getLogger("PYWPS")

# Can be set at runtime with `$ env RAVENPY_THREDDS_URL=https://xx.yy.zz/thredds/ ...`.
THREDDS_URL = os.environ.get("RAVENPY_THREDDS_URL", "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/")
if not THREDDS_URL.endswith("/"):
    THREDDS_URL = f"{THREDDS_URL}/"


def climatology_esp(
    config,
    years: Optional[list[int]] = None,
    workdir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> EnsembleReader:
    """
    Ensemble Streamflow Prediction based on historical variability.

    Run the model using forcing for different years.
    No model warm-up is performed by this function, make sure the initial states are
    consistent with the start date.

    Parameters
    ----------
    config : ravenpy.config.rvs.Config
        Model configuration.
    years : List[int], optional
        Years from which forcing time series will be drawn.
        If None, run for all years where forcing data is available.
    workdir : str or Path
        The path to rv files and model outputs. If None, create a temporary directory.
    overwrite : bool
        Whether to overwrite existing values or not. Default: False.

    Returns
    -------
    EnsembleReader
        Class facilitating the analysis of multiple Raven outputs.
    """
    workdir = Path(workdir or tempfile.mkdtemp())

    # Time from meteo forcing files.
    time = _get_time(config)

    if years is None:
        # Create list of years
        years = list(range(time[0].dt.year.values, time[-1].dt.year.values + 1))

    # Define duration and end_date
    # Make sure configuration uses duration instead of end_date
    if config.duration is None:
        if config.end_date is None or config.start_date is None:
            raise ValueError("StartDate and Duration must be configured.")
        duration = config.end_date - config.start_date
        end = config.end_date
        config.duration = duration.days
        config.end_date = None

    else:
        if config.start_date is None:
            raise ValueError("StartDate must be configured.")
        duration = config.duration
        end = config.start_date + dt.timedelta(days=duration)

    # Define start date
    start = config.start_date

    if start.month == 2 and start.day == 29:
        start = start.replace(day=28)

    # Get the start date that is immutable
    startdate_fixed = deepcopy(start)

    # Run the model for each year
    ensemble = []
    for year in years:
        # Prepare model instance
        config.start_date = s = startdate_fixed.replace(year=year)
        if len(time.sel(time=slice(str(s), str(end.replace(year=year))))) < duration:
            continue

        out = Emulator(config=config, workdir=workdir / f"Y{year}").run(overwrite=overwrite)

        # Format output so that `time` is set to the forecast time, not the forcing time.
        if "hydrograph" in out.files:
            out.files["hydrograph"] = _shift_esp_time(out.files["hydrograph"], startdate_fixed.year)
        if "storage" in out.files:
            out.files["storage"] = _shift_esp_time(out.files["storage"], startdate_fixed.year)
        ensemble.append(out)

    return EnsembleReader(run_name=config.run_name, runs=ensemble)


def to_climpred_hindcast_ensemble(hindcast: xr.Dataset, observations: xr.Dataset) -> climpred.HindcastEnsemble:
    """
    Create a hindcasting object that can be used by the `climpred` toolbox for hindcast verification.

    Parameters
    ----------
    hindcast : xarray.Dataset
        The hindcasted streamflow data for a given period.
    observations : xarray.Dataset
        The streamflow observations that are used to verify the hindcasts.

    Returns
    -------
    climpred.HindcastEnsemble
        The hindcast ensemble formatted to be used in climpred.
    """
    # Make sure the variable names are the same.
    var_names = set(hindcast.data_vars.keys()).intersection(observations.data_vars.keys())
    if len(var_names) == 0:
        raise ValueError("`hindcast` and `observations` should have a variable with the same name.")

    # Todo: Add verification of sizes, catch and return message.

    # Make the hindcastEnsemble object for the hindcast data
    hindcast_obj = HindcastEnsemble(hindcast)

    # Add the observations to that hindcastEnsemble object for verification.
    hindcast_obj = hindcast_obj.add_observations(observations)

    return hindcast_obj


def warm_up(
    config,
    duration: int,
    workdir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> Config:
    """
    Run the model on a time series preceding the start date.

    Parameters
    ----------
    config : ravenpy.config.rvs.Config
        Model configuration.
    duration : int
        Number of days the warm-up simulation should last *before* the start date.
    workdir : Path
        Work directory.
    overwrite : bool
        If True, overwrite existing files.

    Returns
    -------
    ravenpy.config.rvs.Config
        Model configuration with initial state set by running the model prior to the start date.
    """
    wup = config.model_copy(deep=True)
    wup.start_date = config.start_date - dt.timedelta(days=duration)
    wup.duration = duration
    wup.end_date = None

    out = Emulator(wup, workdir=workdir).run(overwrite=overwrite)
    return config.set_solution(out.files["solution"])


def hindcast_climatology_esp(
    config: Config,
    warm_up_duration: int,
    years: Optional[list[int]] = None,
    hindcast_years: Optional[list[int]] = None,
    workdir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> xr.Dataset:
    """
    Hindcast of Ensemble Prediction Streamflow.

    This function runs an emulator initialized for each year in `hindcast_years`,
    using the forcing time series for each year in `years`. This allows an assessment
    of the performance of the ESP. The total number of simulations is given by
    `len(years) * len(hindcasts_years)`.

    Parameters
    ----------
    config : ravenpy.config.rvs.Config
        Model configuration. Initial states will be overwritten.
    warm_up_duration : int
        Number of days to run the model prior to the starting date to initialize the state variables.
    years : List[int]
        Years from which forcing time series will be drawn.
        If None, run for all years where forcing data is available.
    hindcast_years : List[int]
        Years for which the model will be initialized and the `climatology_esp` function run.
        Defaults to all years when forcing data is available.
    workdir : Path
        Work directory. If None, creates a temporary directory.
    overwrite : bool
        If True, overwrite existing files.

    Returns
    -------
    xarray.DataArray
        The array containing the (init, member, lead) dimensions ready for using in climpred (qsim).

    Notes
    -----
    The dataset output dimensions are:
     - `init`: hindcast issue date,
     - `member`: ESP members of the hindcasting experiment,
     - `lead`: number of lead days of the forecast.
    """
    workdir = Path(workdir or tempfile.mkdtemp())

    # `hindcast_years` defaults to all available years
    time = _get_time(config)
    if hindcast_years is None:
        # Create list of years
        hindcast_years = list(range(time[0].dt.year.values, time[-1].dt.year.values + 1))

    # Define start date
    start = config.start_date
    if start.month == 2 and start.day == 29:
        start = start.replace(day=28)

    # Run climatology_esp for each year in `hindcast_years`, with initial conditions set
    # by a warm-up simulation.
    q_sims = {}
    for year in hindcast_years:
        # Compute initial state for each year
        config.start_date = e = start.replace(year=year)
        s = e - dt.timedelta(days=warm_up_duration)
        if len(time.sel(time=slice(str(s), str(e)))) < warm_up_duration:
            continue

        # Run warm-up simulation to set initial state
        conf = warm_up(
            config,
            duration=warm_up_duration,
            workdir=workdir / "wup" / f"Y{year}",
            overwrite=overwrite,
        )

        # Run climatology ESP
        esp = climatology_esp(conf, years=years, workdir=workdir / "esp" / f"Y{year}", overwrite=overwrite)

        # Format results for climpred
        q_sim = esp.hydrograph.q_sim.expand_dims(
            init=[
                config.start_date,
            ]
        )
        q_sim = q_sim.rename({"time": "lead"})
        q_sim.lead.attrs["units"] = "days"
        q_sim = q_sim.assign_coords(lead=list(range(1, len(q_sim.lead) + 1)))

        q_sims[year] = q_sim

    # Format results for climpred
    q_sims = xr.concat(q_sims.values(), dim="init")
    q_sims.lead.attrs["units"] = "days"
    # q_sims = q_sims.assign_coords(init=q_sims.init)
    q_sims = q_sims.to_dataset()

    return q_sims


def _shift_esp_time(nc, year, dim="member"):
    """
    Modify netCDF time to facilitate ensemble analysis out of ESP forecasts.

    Modify time such that it starts on the given year, and add member dimension with original year as value.
    """
    ds = xr.open_dataset(nc)

    # Create new time coordinate
    start = ds.indexes["time"][0]
    freq = xr.infer_freq(ds.time)
    ds["time"] = xr.date_range(
        start.replace(year=year),
        periods=len(ds.time),
        freq=freq,
        calendar=ds.time.dt.calendar,
    )

    # New coordinate dimension to store the original year
    out = ds.expand_dims(
        {
            dim: [
                start.year,
            ]
        }
    )

    # Write to disk
    fn = nc.parent.joinpath(f"{nc.stem}_shifted{nc.suffix}")
    out.to_netcdf(fn, mode="w")
    return fn


def _get_time(config: Config) -> xr.DataArray:
    """Return time DataArray."""
    if config.gauge is not None:
        time = config.gauge[0].ds.time
    elif config.station_forcing is not None:
        time = config.station_forcing[0].da.time
    elif config.gridded_forcing is not None:
        time = config.gridded_forcing[0].da.time
    else:
        raise NotImplementedError()
    return time


def ensemble_prediction(
    config,
    forecast: Union[str, Path],
    ens_dim: str = "member",
    workdir=None,
    overwrite=True,
    **kwds,
) -> EnsembleReader:
    r"""
    Ensemble Streamflow Prediction based on historical weather forecasts (CASPAR or other).

    Run the model using forcing for different years.
    No model warm-up is performed by this function, make sure the initial states are consistent with the start date.

    Parameters
    ----------
    config : ravenpy.config.rvs.Config
        Model configuration.
    forecast : str or Path
        Forecast subsetted to the catchment location (.nc).
    ens_dim : str
        Name of dimension to iterate over.
    workdir : str or Path
        The path to rv files and model outputs. If None, create temporary directory.
    overwrite : bool
        Overwrite files when writing to disk.
    \*\*kwds : dict
        Keywords for the `Gauge.from_nc` function.

    Returns
    -------
    EnsembleReader
        Class facilitating the analysis of multiple Raven outputs.
    """
    workdir = Path(workdir or tempfile.mkdtemp())

    # Run the model for each year
    ensemble = []

    time_coder = xr.coders.CFDatetimeCoder(use_cftime=True)
    forecast_ds = xr.open_dataset(forecast, decode_times=time_coder)

    for member in range(0, len(forecast_ds[ens_dim])):
        # Prepare model instance

        out = Emulator(
            config=config.model_copy(
                update={
                    "gauge": [
                        rc.Gauge.from_nc(
                            forecast,
                            station_idx=member + 1,
                            **kwds,
                        ),
                    ]
                },
            ),
            workdir=workdir / f"Y{member}",
        ).run(overwrite=overwrite)

        # Append to the ensemble.
        ensemble.append(out)

    return EnsembleReader(run_name=config.run_name, runs=ensemble, dim=ens_dim)


# Alias
hindcast_from_meteo_forecast = ensemble_prediction


def compute_forecast_flood_risk(forecast: xr.Dataset, flood_level: float, thredds: str = THREDDS_URL) -> xr.Dataset:
    """
    Return the empirical exceedance probability for each forecast day based on a flood level threshold.

    Parameters
    ----------
    forecast : xr.Dataset
        Ensemble or deterministic streamflow forecast.
    flood_level : float
        Flood level threshold. Will be used to determine if forecasts exceed
        this specified flood threshold. Should be in the same units as the forecasted streamflow.
    thredds : str
        The thredds server url. Default: "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/".

    Returns
    -------
    xr.Dataset
        Time series of probabilities of flood level exceedance.
    """
    if thredds[-1] != "/":
        warnings.warn("The thredds url should end with a slash. Appending it to the url.", stacklevel=2)
        thredds = f"{thredds}/"

    # ---- Calculations ---- #
    # Ensemble: for each day, calculate the percentage of members that are above the threshold
    if "member" in forecast.coords:
        # Get number of members originally
        number_members = len(forecast.member)

        # now compute the ratio of cases that are above the threshold
        pct = forecast.where(forecast > flood_level).notnull().sum(dim="member", keep_attrs=True) / number_members

    # it's deterministic:
    else:
        pct = forecast.where(forecast > flood_level).notnull() / 1.0  # This is needed to return values instead of floats

    domain = urlparse(thredds).netloc

    out = pct.to_dataset(name="exceedance_probability")
    out.attrs["source"] = f"PAVICS-Hydro flood risk forecasting tool, {domain}"
    out.attrs["history"] = (
        f"File created on {dt.datetime.now(dt.UTC).strftime('%Y-%m-%d %H:%M:%S')} UTC on the PAVICS-Hydro service available at {domain}."
    )
    out.attrs["title"] = "Identification of ensemble members that exceed a certain flow threshold."

    return out



"""Configurations for raven-hydro."""

import os
import shutil


RAVEN_EXEC_PATH = os.getenv("RAVENPY_RAVEN_BINARY_PATH") or shutil.which("raven")

if not RAVEN_EXEC_PATH:
    raise RuntimeError("Could not find raven binary in PATH and RAVENPY_RAVEN_BINARY_PATH env variable is not set.")

try:
    from raven_hydro import __raven_version__
except ImportError:
    __raven_version__ = "0.0.0"



"""Tools for hydrological regionalization."""

import logging
import tempfile
from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
import pandas as pd
import statsmodels.api as sm
import xarray as xr
from haversine import haversine_vector

from ravenpy.config import Config
from ravenpy.ravenpy import Emulator, EnsembleReader

from . import coords


# from statsmodels.regression.linear_model import RegressionResults


LOGGER = logging.getLogger("PYWPS")

regionalisation_data_dir = Path(__file__).parent.parent / "data" / "regionalisation"


def regionalize(
    config: Config,
    method: str,
    nash: pd.Series,
    params: pd.DataFrame = None,
    props: pd.DataFrame = None,
    target_props: Union[pd.Series, dict] = None,
    size: int = 5,
    min_NSE: float = 0.6,  # noqa: N803
    workdir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> tuple[xr.DataArray, xr.Dataset]:
    r"""
    Perform regionalization for catchment whose outlet is defined by coordinates.

    Parameters
    ----------
    config : ravenpy.config.rvs.Config
        Symbolic emulator configuration. Only GR4JCN, HMETS and Mohyse are supported.
    method : {'MLR', 'SP', 'PS', 'SP_IDW', 'PS_IDW', 'SP_IDW_RA', 'PS_IDW_RA'}
        Name of the regionalization method to use.
    nash : pd.Series
        NSE values for the parameters of gauged catchments.
    params : pd.DataFrame
        Model parameters of gauged catchments. Needed for all but MRL method.
    props : pd.DataFrame
        Properties of gauged catchments to be analyzed for the regionalization.
        Needed for MLR and RA methods.
    target_props : pd.Series or dict
        Properties of ungauged catchment. Needed for MLR and RA methods.
    size : int
        Number of catchments to use in the regionalization.
    min_NSE : float
        Minimum calibration NSE value required to be considered as a donor.
    workdir : Union[str, Path]
        Work directory. If None, a temporary directory will be created.
    overwrite : bool
        If True, existing files will be overwritten.

    Returns
    -------
    qsim : DataArray (time, )
        Multi-donor averaged predicted streamflow.
    ensemble : Dataset
        A Dataset containing the ensemble of simulations and parameters used:

        - q_sim : DataArray (realization, time)
          Ensemble of members based on number of donors.

        - parameter : DataArray (realization, param)
          Parameters used to run the model.
    """
    name = config.__class__.__name__
    if name not in ["GR4JCN", "HMETS", "Mohyse"]:
        raise ValueError(f"Emulator {name} is not supported for regionalization.")

    if not config.is_symbolic:
        raise ValueError("config should be a symbolic configuration.")

    # TODO: Include list of available properties in docstring.
    # TODO: Add error checking for source, target stuff wrt method chosen.
    workdir = Path(workdir or tempfile.mkdtemp())

    # Select properties based on those available in the ungauged properties DataFrame.
    if isinstance(target_props, dict):
        ungauged_properties = pd.Series(target_props)
    elif isinstance(target_props, pd.Series):
        ungauged_properties = target_props
    elif isinstance(target_props, pd.DataFrame):
        ungauged_properties = target_props.to_series()
    else:
        raise ValueError

    cr = coords.realization(1 if method == "MLR" else size)
    cp = coords.param(name)

    # Filter on NSE
    valid = nash > min_NSE
    filtered_params = params.where(valid).dropna()
    filtered_prop = props.where(valid).dropna()

    # Check to see if we have enough data, otherwise raise error
    if len(filtered_prop) < size and method != "MLR":
        raise ValueError(
            "Hydrological_model and minimum NSE threshold \
                         combination is too strict for the number of donor \
                         basins. Please reduce the number of donor basins OR \
                         reduce the minimum NSE threshold."
        )

    # Rank the matrix according to the similarity or distance.
    if method in ["PS", "PS_IDW", "PS_IDW_RA"]:  # Physical similarity
        dist = similarity(filtered_prop, ungauged_properties)
    else:  # Geographical distance.
        dist = distance(filtered_prop, ungauged_properties)

    # Series of distances for the first `size` best donors
    sdist = dist.sort_values().iloc[:size]

    # Pick the donors' model parameters and catchment properties
    sparams = filtered_params.loc[sdist.index]
    sprop = filtered_prop.loc[sdist.index]

    # Get the list of parameters to run
    reg_params = regionalization_params(method, sparams, sprop, ungauged_properties, filtered_params, filtered_prop)

    # Run the model over all parameters and create ensemble DataArray

    ensemble = []
    for i, rparams in enumerate(reg_params):
        model_config_tmp = config.set_params(rparams)

        out = Emulator(model_config_tmp, workdir=workdir / f"donor_{i}").run(overwrite=overwrite)

        # Append to the ensemble.
        ensemble.append(out)

    qsim_obj = EnsembleReader(runs=ensemble, dim="members")
    qsims = qsim_obj.hydrograph.q_sim

    # 3. Aggregate runs into a single result -> dataset
    if method in [
        "MLR",
        "SP",
        "PS",
    ]:  # Average (one realization for MLR, so no effect).
        qsim = qsims.mean(dim="members", keep_attrs=True)
    elif "IDW" in method:  # Here we are replacing the mean by the IDW average, keeping attributes and dimensions.
        qsim = IDW(qsims, sdist)
    else:
        raise ValueError(f"No matching algorithm for {method}")

    # Metadata handling
    # TODO: Store the basin_name

    # Create a DataArray for the parameters used in the regionalization
    param_da = xr.DataArray(
        reg_params,
        dims=("members", "param"),
        coords={"param": cp, "members": cr},
        attrs={"long_name": "Model parameters used in the regionalization."},
    )

    ens = xr.Dataset(
        data_vars={"q_sim": qsims, "parameter": param_da},
        attrs={
            "title": "Regionalization ensemble",
            # "institution": "",
            "source": f"Hydrological model {name}",
            "history": "Created by ravenpy regionalize.",
            # "references": "",
            "comment": f"Regionalization method: {method}",
        },
    )

    # TODO: Add global attributes (model name, date, version, etc)
    return qsim, ens


def read_gauged_properties(properties) -> pd.DataFrame:
    """
    Return table of gauged catchments properties over North America.

    Returns
    -------
    pd.DataFrame
        Catchment properties keyed by catchment ID.
    """
    f = regionalisation_data_dir / "gauged_catchment_properties.csv"
    proptable = pd.read_csv(f, index_col="ID")

    return proptable[properties]


def read_gauged_params(model):
    """
    Return table of NASH-Sutcliffe Efficiency values and model parameters for North American catchments.

    Returns
    -------
    pd.DataFrame
        Nash-Sutcliffe Efficiency keyed by catchment ID.
    pd.DataFrame
        Model parameters keyed by catchment ID.
    """
    f = regionalisation_data_dir / f"{model}_parameters.csv"
    params = pd.read_csv(f, index_col="ID")

    return params["NASH"], params.iloc[:, 1:]


def distance(gauged: pd.DataFrame, ungauged: pd.Series) -> pd.Series:
    """
    Return geographic distance [km] between ungauged and database of gauged catchments.

    Parameters
    ----------
    gauged : pd.DataFrame
        Table containing columns for longitude and latitude of catchment's centroid.
    ungauged : pd.Series
        Coordinates of the ungauged catchment.

    Returns
    -------
    pd.Series
    """
    gauged_array = np.array(list(zip(gauged.latitude.values, gauged.longitude.values)))

    return pd.Series(
        data=haversine_vector(gauged_array, np.array([ungauged.latitude, ungauged.longitude]), comb=True)[0],
        index=gauged.index,
    )


def similarity(gauged: pd.DataFrame, ungauged: pd.DataFrame, kind: str = "ptp") -> pd.Series:
    """
    Return similarity measure between gauged and ungauged catchments.

    Parameters
    ----------
    gauged : pd.DataFrame
        Gauged catchment properties.
    ungauged : pd.DataFrame
        Ungauged catchment properties.
    kind : {'ptp', 'std', 'iqr'}
        Normalization method: peak to peak (maximum - minimum), standard deviation, inter-quartile range.

    Returns
    -------
    pd.Series
    """
    stats = gauged.describe()

    if kind == "ptp":
        spread = stats.loc["max"] - stats.loc["min"]
    elif kind == "std":
        spread = stats.loc["std"]
    elif kind == "iqr":
        spread = stats.loc["75%"] - stats.loc["25%"]

    d = ungauged.values - gauged.values
    n = np.abs(d) / spread.values
    return pd.Series(data=n.sum(axis=1), index=gauged.index)


# FIXME: gauged_properties is not used
def regionalization_params(
    method: str,
    gauged_params: pd.DataFrame,
    gauged_properties: pd.DataFrame,  # noqa: F841
    ungauged_properties: pd.DataFrame,
    filtered_params: pd.DataFrame,
    filtered_prop: pd.DataFrame,
) -> Union[list[list[float]], list[np.ndarray]]:
    """
    Return the model parameters to use for the regionalization.

    Parameters
    ----------
    method : {'MLR', 'SP', 'PS', 'SP_IDW', 'PS_IDW', 'SP_IDW_RA', 'PS_IDW_RA'}
        Name of the regionalization method to use.
    gauged_params : pd.DataFrame
        A DataFrame of parameters for donor catchments (size = number of donors).
    gauged_properties : pd.DataFrame
        A DataFrame of properties of the donor catchments  (size = number of donors).
    ungauged_properties : pd.DataFrame
        A DataFrame of properties of the ungauged catchment (size = 1).
    filtered_params : pd.DataFrame
        A DataFrame of parameters of all filtered catchments (size = all catchments with NSE > min_NSE).
    filtered_prop : pd.DataFrame
        A DataFrame of properties of all filtered catchments (size = all catchments with NSE > min_NSE).

    Returns
    -------
    list
        A list of model parameters to be used for the regionalization.
    """
    if method == "MLR" or "RA" in method:
        mlr_params, r2 = multiple_linear_regression(filtered_prop, filtered_params, ungauged_properties.to_frame().T)

        if method == "MLR":  # Return the multiple linear regression parameters.
            out = [
                mlr_params,
            ]

        else:
            gp = gauged_params.copy()

            for p, r, col in zip(mlr_params, r2, gauged_params):
                # If we have an R2 > 0.5 then we consider this to be a better estimator

                if r > 0.5:
                    gp[col] = p

            out = gp.values

    else:
        out = gauged_params.values

    return out


def IDW(qsims: xr.DataArray, dist: pd.Series) -> xr.DataArray:  # noqa: N802
    """
    Inverse distance weighting.

    Parameters
    ----------
    qsims : xr.DataArray
        Ensemble of hydrogram stacked along the `members` dimension.
    dist : pd.Series
        Distance from catchment which generated each hydrogram to target catchment.

    Returns
    -------
    xr.DataArray
        Inverse distance weighted average of ensemble.
    """
    # In IDW, weights are 1 / distance
    weights = xr.DataArray(1.0 / dist, dims="members", coords={"members": qsims.members})

    # Make weights sum to one
    weights /= weights.sum(axis=0)

    # Calculate weighted average.
    out = qsims.dot(weights)
    out.name = qsims.name
    out.attrs = qsims.attrs
    return out


def multiple_linear_regression(source: pd.DataFrame, params: pd.DataFrame, target: pd.DataFrame) -> tuple[list[Any], list[Callable[[], Any]]]:
    """
    Multiple Linear Regression for model parameters over catchment properties.

    Uses known catchment properties and model parameters to estimate model parameter over an
    ungauged catchment using its properties.

    Parameters
    ----------
    source : pd.DataFrame
        Properties of gauged catchments.
    params : pd.DataFrame
        Model parameters of gauged catchments.
    target : pd.DataFrame
        Properties of the ungauged catchment.

    Returns
    -------
    list of Any, list of Callable or Any
        A named tuple of the estimated model parameters and the R2 of the linear regression.
    """
    # Add constants to the gauged predictors
    x = sm.add_constant(source)

    # Add the constant '1' for the ungauged catchment predictors
    predictors = sm.add_constant(target, prepend=True, has_constant="add")

    # Perform regression for each parameter
    regression = [sm.OLS(params[param].values, x).fit() for param in params]

    # Perform prediction on each parameter based on the predictors
    mlr_parameters = [r.predict(exog=predictors)[0] for r in regression]

    # Extract the adjusted r_squared value for each parameter
    r2 = [r.rsquared_adj for r in regression]

    return mlr_parameters, r2



"""Tools for performing geospatial translations and transformations."""

import collections
import logging
from pathlib import Path
from typing import Optional, Union

import geopandas
import pandas as pd

from . import gis_import_error_message


try:
    import fiona
    import rasterio
    import rasterio.mask
    import rasterio.vrt
    import rasterio.warp
    from pyproj import CRS
    from shapely.geometry import (
        GeometryCollection,
        MultiPolygon,
        Point,
        Polygon,
        mapping,
        shape,
    )
    from shapely.ops import transform
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

RASTERIO_TIFF_COMPRESSION = "lzw"
LOGGER = logging.getLogger("RavenPy")
WGS84 = 4326


def geom_transform(
    geom: Union[GeometryCollection, shape],
    source_crs: Union[str, int, CRS] = WGS84,
    target_crs: Union[str, int, CRS] = None,
) -> GeometryCollection:
    """
    Change the projection of a geometry.

    Assuming a geometry's coordinates are in a `source_crs`, compute the new coordinates under the `target_crs`.

    Parameters
    ----------
    geom : Union[GeometryCollection, shape]
        Source geometry.
    source_crs : Union[str, int, CRS]
        Projection identifier (proj4) for the source geometry, e.g. '+proj=longlat +datum=WGS84 +no_defs'.
    target_crs : Union[str, int, CRS]
        Projection identifier (proj4) for the target geometry.

    Returns
    -------
    GeometryCollection
        Reprojected geometry.
    """
    try:
        from pyproj import Transformer

        if isinstance(source_crs, int or str):
            source = CRS.from_epsg(source_crs)
        else:
            source = source_crs

        if isinstance(target_crs, int or str):
            target = CRS.from_epsg(target_crs)
        else:
            target = target_crs

        transform_func = Transformer.from_crs(source, target, always_xy=True)
        reprojected = transform(transform_func.transform, geom)

        return reprojected
    except Exception as err:  # noqa: BLE001
        msg = f"{err}: Failed to reproject geometry"
        LOGGER.error(msg)
        raise Exception(msg) from err


def generic_raster_clip(
    raster: Union[str, Path],
    output: Union[str, Path],
    geometry: Union[Polygon, MultiPolygon, list[Union[Polygon, MultiPolygon]]],
    touches: bool = False,
    fill_with_nodata: bool = True,
    padded: bool = True,
    raster_compression: str = RASTERIO_TIFF_COMPRESSION,
) -> None:
    """
    Crop a raster file to a given geometry.

    Parameters
    ----------
    raster : Union[str, Path]
        Path to input raster.
    output : Union[str, Path]
        Path to output raster.
    geometry : Union[Polygon, MultiPolygon, List[Union[Polygon, MultiPolygon]]
        Geometry defining the region to crop.
    touches : bool
        Whether to include cells that intersect the geometry or not. Default: True.
    fill_with_nodata : bool
        Whether to keep pixel values for regions outside of shape or set as nodata or not. Default: True.
    padded : bool
        Whether to add a half-pixel buffer to shape before masking or not. Default: True.
    raster_compression : str
        Level of data compression. Default: 'lzw'.

    Returns
    -------
    None
    """
    if not isinstance(geometry, collections.abc.Iterable):
        geometry = [geometry]

    with rasterio.open(raster, "r") as src:
        mask_image, mask_affine = rasterio.mask.mask(
            src,
            geometry,
            crop=True,
            pad=padded,
            all_touched=touches,
            filled=fill_with_nodata,
        )
        mask_meta = src.meta.copy()
        mask_meta.update(
            {
                "driver": "GTiff",
                "height": mask_image.shape[1],
                "width": mask_image.shape[2],
                "transform": mask_affine,
                "compress": raster_compression,
            }
        )

        # Write the new masked image
        with rasterio.open(output, "w", **mask_meta) as dst:
            dst.write(mask_image)


def generic_raster_warp(
    raster: Union[str, Path],
    output: Union[str, Path],
    target_crs: Union[str, dict, CRS],
    raster_compression: str = RASTERIO_TIFF_COMPRESSION,
) -> None:
    """
    Reproject a raster file.

    Parameters
    ----------
    raster : Union[str, Path]
        Path to input raster.
    output : Union[str, Path]
        Path to output raster.
    target_crs : str or dict
        Target projection identifier.
    raster_compression : str
        Level of data compression. Default: 'lzw'.

    Returns
    -------
    None
    """
    with rasterio.open(raster, "r") as src:
        # Reproject raster using WarpedVRT class
        with rasterio.vrt.WarpedVRT(src, crs=target_crs) as vrt:
            # Calculate grid properties based on projection
            affine, width, height = rasterio.warp.calculate_default_transform(src.crs, target_crs, src.width, src.height, *src.bounds)

            # Copy relevant metadata from parent raster
            metadata = src.meta.copy()
            metadata.update(
                {
                    "driver": "GTiff",
                    "height": height,
                    "width": width,
                    "transform": affine,
                    "crs": target_crs,
                    "compress": raster_compression,
                }
            )
            data = vrt.read()

            with rasterio.open(output, "w", **metadata) as dst:
                dst.write(data)


def generic_vector_reproject(
    vector: Union[str, Path],
    projected: Union[str, Path],
    source_crs: Union[str, CRS] = WGS84,
    target_crs: Union[str, CRS] = None,
) -> None:
    """
    Reproject all features and layers within a vector file and return a GeoJSON.

    Parameters
    ----------
    vector : Union[str, Path]
        Path to a file containing a valid vector layer.
    projected : Union[str, Path]
        Path to a file to be written.
    source_crs : Union[str, pyproj.crs.CRS]
        CRS for the source geometry. Default: 4326.
    target_crs : Union[str, pyproj.crs.CRS]
        CRS for the target geometry.

    Returns
    -------
    None
    """
    if target_crs is None:
        raise ValueError("No target CRS is defined.")
    if isinstance(target_crs, CRS):
        target_crs = target_crs.to_dict()

    if isinstance(vector, Path):
        vector = vector.as_posix()

    for i, _ in enumerate(fiona.listlayers(vector)):
        with fiona.open(vector, "r", layer=i) as src:
            with fiona.open(projected, "w", driver="GeoJSON", schema=src.schema, crs=target_crs) as sink:
                for feature in src:
                    # Perform vector reprojection using Shapely on each feature
                    try:
                        geom = shape(feature["geometry"])
                        transformed = geom_transform(geom, source_crs, target_crs)
                        transformed_feature = {
                            "geometry": fiona.Geometry().from_dict(mapping(transformed)),
                            "properties": feature.__geo_interface__["properties"],
                            "id": feature.__geo_interface__["id"],
                            "type": feature.__geo_interface__["type"],
                        }
                        sink.write(transformed_feature)
                    except Exception as err:  # noqa: PERF203
                        msg = f"{err}: Unable to reproject feature {feature}"
                        LOGGER.exception(msg)
                        raise


def determine_upstream_ids(
    fid: Union[str, int, float],
    df: Union[pd.DataFrame, geopandas.GeoDataFrame],
    *,
    basin_field: str,
    downstream_field: str,
    basin_family: Optional[str] = None,
) -> Union[pd.DataFrame, geopandas.GeoDataFrame]:
    """
    Return a list of upstream features by evaluating the downstream networks.

    Parameters
    ----------
    fid : str or int or float
        The feature ID of the downstream feature of interest.
    df : pd.DataFrame
        A Dataframe comprising the watershed attributes.
    basin_field : str
        The field used to determine the id of the basin according to hydro project.
    downstream_field : str
        The field identifying the downstream sub-basin for the hydro project.
    basin_family : str, optional
        Regional watershed code (For HydroBASINS dataset).

    Returns
    -------
    pd.DataFrame
        Basins ids including `fid` and its upstream contributors.
    """

    def upstream_ids(bdf, bid):
        return bdf[bdf[downstream_field] == bid][basin_field]

    # Note: Hydro Routing `SubId` is a float for some reason and Python float != GeoServer double. Cast them to int.
    if isinstance(fid, float):
        fid = int(fid)
        df[basin_field] = df[basin_field].astype(int)
        df[downstream_field] = df[downstream_field].astype(int)

    # Locate the downstream feature
    ds = df.set_index(basin_field).loc[fid]
    if basin_family is not None:
        # Do a first selection on the main basin ID of the downstream feature.
        sub = df[df[basin_family] == ds[basin_family]]
    else:
        sub = None

    # Find upstream basins
    up = [fid]
    for b in up:
        tmp = upstream_ids(sub if sub is not None else df, b)
        if len(tmp):
            up.extend(tmp)  # noqa: B909

    return sub[sub[basin_field].isin(up)] if sub is not None else df[df[basin_field].isin(up)]


def find_geometry_from_coord(lon: float, lat: float, df: geopandas.GeoDataFrame) -> geopandas.GeoDataFrame:
    """
    Return the geometry containing the given coordinates.

    lon : float
        Longitude.
    lat : float
        Latitude.
    df : GeoDataFrame
        Data.

    Returns
    -------
    GeoDataFrame
        Record whose geometry contains the point.
    """
    p = Point(lon, lat)

    c = df.contains(p)
    n = c.sum()
    if n == 0:
        raise ValueError(f"Point {p} not in any geometry.")
    elif n > 1:
        raise ValueError(f"Point {p} found in multiple geometries.")

    return df[c]



from ravenpy.config import commands as rc


def test_line_command():
    c = rc.SBGroupPropertyMultiplier(group_name="Land", parameter_name="MANNINGS_N", mult=1.0)

    out = c.to_rv()
    assert out == ":SBGroupPropertyMultiplier Land MANNINGS_N 1.0\n"



from collections.abc import Sequence
from typing import Literal, Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Params, Sym, SymConfig
from ravenpy.config.commands import HRU, PL, Process
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


@dataclass(config=SymConfig)
class P(Params):
    """
    Model parameters.

    Default parameters values are mainly taken from the Raven RavenParameters.dat.
    Min and max values are taken either from the Raven RavenParameters.dat or the manual (v3.7),
    when both sources differed or when values were missing (value of -9999) in the RavenParameters.dat file.
    """

    X01: Sym = Variable(
        "X01"
    )  # RAINSNOW_TEMP      | rain/snow halfway transition temperature [◦C]                  | default=-0.15, min=-3.0, max=3.0
    X02: Sym = Variable(
        "X02"
    )  # MELT_FACTOR        | maximum snow melt factor used in degree day models [mm/d/◦C]   | default=5.04, min=-9999, max=-9999
    X03: Sym = Variable(
        "X03"
    )  # REFREEZE_FACTOR    | maximum refreeze factor used in degree day models [mm/d/◦C]    | default=5.04, min=0.0, max=10.0
    X04: Sym = Variable(
        "X04"
    )  # SNOW_SWI           | water saturation fraction of snow [-]                          | default=0.05, min=0.04, max=0.07
    X05: Sym = Variable("X05")  # POROSITY           | effective porosity of the soil [-]                             | default=0.4, min=0.0, max=1.0
    X06: Sym = Variable("X06")  # FIELD_CAPACITY     | field capacity saturation of the soil [-]                      | default=0.0, min=0.0, max=1.0
    X07: Sym = Variable("X07")  # HBV_BETA           | HBV infiltration exponent [-]                                  | default=1.0, min=0.0, max=7.0
    X08: Sym = Variable(
        "X08"
    )  # MAX_PERC_RATE      | VIC/ARNO/GAWSER percolation rate [mm/d]                        | default=-9999, min=0.01, max=1000
    X09: Sym = Variable("X09")  # BASEFLOW_COEFF     | linear baseflow storage/routing coeff (FAST_RES) [1/d]         | default=0.1, min=0.0, max=9999
    X10: Sym = Variable("X10")  # BASEFLOW_COEFF     | linear baseflow storage/routing coeff (SLOW_RES) [1/d]         | default=0.1, min=0.0, max=9999
    X11: Sym = Variable(
        "X11"
    )  # TIME_CONC          | the time of concentration of the unit hydrograph [d]           | default=1.0, min=0.0, max=200.0
    X12: Sym = Variable(
        "X12"
    )  # PRECIP_LAPSE       | precipitation lapse rate for orographic correction [mm/d/km]   | default=-9999, min=0.0, max=100.0
    X13: Sym = Variable("X13")  # ADIABATIC_LAPSE    | adiabatic temperature lapse rate [◦C/km]                       | default=6.4, min=5.0, max=8.0
    X14: Sym = Variable("X14")  # SAT_WILT           | hydroscopic minimum saturation [-]                             | default=0.0, min=0.0, max=1.0
    X15: Sym = Variable("X15")  # BASEFLOW_N         | VIC/ARNO baseflow exponent [-]                                 | default=5.0, min=1.0, max=10.0
    X16: Sym = Variable(
        "X16"
    )  # MAX_CAP_RISE_RATE  | HBV max capillary rise rate [mm/d]                             | default=0.0, min=-9999, max=-9999
    X17: Sym = Variable(
        "X17"
    )  # TOPSOIL THICKNESS  | soil layer thickness (TOPSOIL) [m]                             | default=-9999, min=0.0, max=-9999
    X18: Sym = Variable("X18")  # HBV_MELT_FOR_CORR  | HBV snowmelt forest correction (MRF in HBV-EC) [-]             | default=1.0, min=0.1, max=1.0
    X19: Sym = Variable("X19")  # GLAC_STORAGE_COEFF | maximum linear storage coefficient for glacial melt [-]        | default=0.1, min=0.0, max=1.0
    X20: Sym = Variable("X20")  # RAIN_CORR          | rain correction factor for subbasin (multiplier) [-]           | default=1.0, min=0.5, max=2.0
    X21: Sym = Variable("X21")  # SNOW_CORR          | snow correction factor for subbasin (multiplier) [-]           | default=1.0, min=0.5, max=2.0


class LandHRU(HRU):
    land_use_class: str = "LU_ALL"
    veg_class: str = "VEG_ALL"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["land"] = "land"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[LandHRU]


class HBVEC(Config):
    """
    Hydrologiska Byråns Vattenbalansavdelning – Environment Canada (HBV-EC).

    References
    ----------
    Lindström, G., et al., 1997. Development and test of the distributed HBV-96 hydrological model.
    Journal of Hydrology, 201 (1–4), 272–288. doi:10.1016/S0022-1694(97)00041-3

    Hamilton, A.S., Hutchinson, D.G., and Moore, R.D., 2000. Estimating winter streamflow using
    conceptual streamflow model. Journal of Cold Regions Engineering, 14 (4), 158–175.
    doi:10.1061/(ASCE)0887-381X(2000)14:4(158)

    Canadian Hydraulics Centre, 2010. Green kenue reference manual.
    Ottawa, Ontario: National Research Council.
    """

    params: P = P()
    hrus: HRUs = Field([LandHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "HBVEC"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    interpolation: o.Interpolation = Field("INTERP_NEAREST_NEIGHBOR", alias="Interpolation")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("TRIANGULAR_UH", alias="CatchmentRoute")
    evaporation: o.Evaporation = Field("PET_FROMMONTHLY", alias="Evaporation")
    ow_evaporation: o.Evaporation = Field("PET_FROMMONTHLY", alias="OW_Evaporation")
    sw_radiation_method: o.SWRadiationMethod = Field("SW_RAD_DEFAULT", alias="SWRadiationMethod")
    sw_cloud_correct: o.SWCloudCorrect = Field("SW_CLOUD_CORR_NONE", alias="SWCloudCorrect")
    sw_canopy_correct: o.SWCanopyCorrect = Field("SW_CANOPY_CORR_NONE", alias="SWCanopyCorrect")
    lw_radiation_method: o.LWRadiationMethod = Field("LW_RAD_DEFAULT", alias="LWRadiationMethod")
    rain_snow_fraction: o.RainSnowFraction = Field("RAINSNOW_HBV", alias="RainSnowFraction")
    potential_melt_method: o.PotentialMeltMethod = Field("POTMELT_HBV", alias="PotentialMeltMethod")
    oro_temp_correct: o.OroTempCorrect = Field("OROCORR_HBV", alias="OroTempCorrect")
    oro_precip_correct: o.OroPrecipCorrect = Field("OROCORR_HBV", alias="OroPrecipCorrect")
    oro_pet_correct: o.OroPETCorrect = Field("OROCORR_HBV", alias="OroPETCorrect")
    cloud_cover_method: o.CloudCoverMethod = Field("CLOUDCOV_NONE", alias="CloudCoverMethod")
    precip_icept_frac: o.PrecipIceptFract = Field("PRECIP_ICEPT_USER", alias="PrecipIcepFract")
    monthly_interpolation_method: o.MonthlyInterpolationMethod = Field("MONTHINT_LINEAR_21", alias="MonthlyInterpolationMethod")
    soil_model: rc.SoilModel = Field(3, alias="SoilModel")
    lake_storage: o.StateVariables = Field("SOIL[2]", alias="LakeStorage")
    hydrologic_processes: Sequence[Union[Process, p.Conditional]] = Field(
        [
            p.SnowRefreeze(algo="FREEZE_DEGREE_DAY", source="SNOW_LIQ", to="SNOW"),
            p.Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.CanopyEvaporation(algo="CANEVP_ALL", source="CANOPY", to="ATMOSPHERE"),
            p.CanopySublimation(algo="CANEVP_ALL", source="CANOPY_SNOW", to="ATMOSPHERE"),
            p.SnowBalance(algo="SNOBAL_SIMPLE_MELT", source="SNOW", to="SNOW_LIQ"),
            p.Overflow(algo="RAVEN_DEFAULT", source="SNOW_LIQ", to="PONDED_WATER"),
            p.Flush(algo="RAVEN_DEFAULT", source="PONDED_WATER", to="GLACIER"),
            p.Conditional(kind="HRU_TYPE", op="IS", value="GLACIER"),
            p.GlacierMelt(algo="GMELT_HBV", source="GLACIER_ICE", to="GLACIER"),
            p.GlacierRelease(algo="GRELEASE_HBV_EC", source="GLACIER", to="SURFACE_WATER"),
            p.Infiltration(algo="INF_HBV", source="PONDED_WATER", to="MULTIPLE"),
            p.Flush(algo="RAVEN_DEFAULT", source="SURFACE_WATER", to="SOIL[1]"),
            p.Conditional(kind="HRU_TYPE", op="IS_NOT", value="GLACIER"),
            p.SoilEvaporation(algo="SOILEVAP_HBV", source="SOIL[0]", to="ATMOSPHERE"),
            p.CapillaryRise(algo="CRISE_HBV", source="SOIL[1]", to="SOIL[0]"),
            p.LakeEvaporation(algo="LAKE_EVAP_BASIC", source="SOIL[2]", to="ATMOSPHERE"),
            p.Percolation(algo="PERC_CONSTANT", source="SOIL[1]", to="SOIL[2]"),
            p.Baseflow(algo="BASE_POWER_LAW", source="SOIL[1]", to="SURFACE_WATER"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[2]", to="SURFACE_WATER"),
        ]
    )

    sub_basin_properties: rc.SubBasinProperties = Field(
        {
            "parameters": ["TIME_CONC", "TIME_TO_PEAK"],
            "records": [{"sb_id": 1, "values": (P.X11, P.X11 / 2)}],
        },
        alias="SubBasinProperties",
    )

    rain_snow_transition: rc.RainSnowTransition = Field({"temp": P.X01, "delta": 2}, alias="RainSnowTransition")
    global_parameter: dict = Field(
        {
            "ADIABATIC_LAPSE": P.X13,
            "SNOW_SWI": P.X04,
            "PRECIP_LAPSE": P.X12,
        },
        alias="GlobalParameters",
    )
    soil_classes: rc.SoilClasses = Field(
        [
            {"name": "TOPSOIL", "mineral": (1, 0, 0), "organic": 0},
            {"name": "FAST_RES", "mineral": (1, 0, 0), "organic": 0},
            {"name": "SLOW_RES", "mineral": (1, 0, 0), "organic": 0},
        ],
        alias="SoilClasses",
    )
    soil_parameter_list: rc.SoilParameterList = Field(
        {
            "parameters": [
                "POROSITY",
                "FIELD_CAPACITY",
                "SAT_WILT",
                "HBV_BETA",
                "MAX_CAP_RISE_RATE",
                "MAX_PERC_RATE",
                "BASEFLOW_COEFF",
                "BASEFLOW_N",
            ],
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(P.X05, P.X06, P.X14, P.X07, P.X16, 0.0, 0.1, 1.0),
                ),
                PL(
                    name="TOPSOIL",
                    values=(P.X05, P.X06, P.X14, P.X07, P.X16, 0.0, 0.1, 1.0),
                ),
                PL(
                    name="FAST_RES",
                    values=(
                        "_DEFAULT",
                        "_DEFAULT",
                        0.0,
                        "_DEFAULT",
                        "_DEFAULT",
                        P.X08,
                        P.X09,
                        P.X15 + 1,
                    ),
                ),
                PL(
                    name="SLOW_RES",
                    values=(
                        "_DEFAULT",
                        "_DEFAULT",
                        0.0,
                        "_DEFAULT",
                        "_DEFAULT",
                        "_DEFAULT",
                        P.X10,
                        1.0,
                    ),
                ),
            ],
        },
        alias="SoilParameterList",
    )
    soil_profiles: rc.SoilProfiles = Field(
        [
            {
                "name": "DEFAULT_P",
                "soil_classes": ["TOPSOIL", "FAST_RES", "SLOW_RES"],
                "thicknesses": (P.X17, 100, 100),
            }
        ],
        alias="SoilProfiles",
    )
    vegetation_classes: rc.VegetationClasses = Field(
        [{"name": "VEG_ALL", "max_ht": 25, "max_lai": 6, "max_leaf_cond": 5.3}],
        alias="VegetationClasses",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": ["MAX_CAPACITY", "MAX_SNOW_CAPACITY", "TFRAIN", "TFSNOW"],
            "pl": [PL(name="VEG_ALL", values=(10000, 10000, 0.88, 0.88))],
        },
        alias="VegetationParameterList",
    )
    land_use_classes: rc.LandUseClasses = Field(
        [{"name": "LU_ALL", "impermeable_frac": 0, "forest_coverage": 1}],
        alias="LandUseClasses",
    )
    land_use_parameter_list: rc.LandUseParameterList = Field(
        {
            "parameters": [
                "MELT_FACTOR",
                "MIN_MELT_FACTOR",
                "HBV_MELT_FOR_CORR",
                "REFREEZE_FACTOR",
                "HBV_MELT_ASP_CORR",
                "HBV_MELT_GLACIER_CORR",
                "HBV_GLACIER_KMIN",
                "GLAC_STORAGE_COEFF",
                "HBV_GLACIER_AG",
            ],
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(P.X02, 2.2, P.X18, P.X03, 0.48, 1.64, 0.05, P.X19, 0.05),
                )
            ],
        },
        alias="LandUseParameterList",
    )

    hru_state_variable_table: rc.HRUStateVariableTable = Field([{"hru_id": 1, "data": {"SOIL[2]": 0.50657}}])
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)

    def __init__(self, **data):
        super().__init__(**data)

        if self.gauge:
            for gauge in self.gauge:
                gauge.rain_correction = self.params.X20
                gauge.snow_correction = self.params.X21



import datetime as dt

import numpy as np
import pytest
import xarray as xr

from ravenpy import Emulator
from ravenpy.config import commands as rc
from ravenpy.config.emulators import GR4JCN, BasicRoute
from ravenpy.config.utils import get_average_annual_runoff


# Expected NSE for emulator configuration from the `config_rv` test fixture.
NSE = {
    "GR4JCN": -0.117301,
    "HMETS": -3.0132,
    "Mohyse": 0.194612,
    "HBVEC": 0.0186633,
    # "CanadianShield": 0.39602, <- This is the original value for CanadianShield with RavenHydroFramework v3.6
    # "CanadianShield": 0.3968, <- This is the value for CanadianShield with RavenHydroFramework v3.7
    "CanadianShield": 0.4001,  # <- This is the new value for CanadianShield with RavenHydroFramework v3.8 and v3.8.1
    "HYPR": 0.685188,
    "SACSMA": -0.0382907,
    # "Blended": -0.913785, <- This is the original value for Blended with RavenHydroFramework v3.8.1
    "Blended": -1.1507,  # <- This is the new value for Blended with RavenHydroFramework v4.0.1
}


def test_rv(config_rv):
    """Test the model configuration can be written to disk."""
    name, path = config_rv
    assert (path / "test.rvi").exists()


def test_run(numeric_config, tmp_path):
    """Test that the emulator actually runs and returns the expected NSE."""
    name, conf = numeric_config

    # assert conf.__config__.allow_mutation

    e = Emulator(config=conf, workdir=tmp_path)
    out = e.run()
    d = out.diagnostics

    np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], NSE[name], 4)

    if name == "CanadianShield":
        pytest.skip("Missing solution due to SuppressOutput.")

    # Start new simulation with final state from initial run.
    new = e.resume()
    assert new.start_date == conf.end_date
    new.end_date = dt.datetime(2002, 1, 7)
    new.run_name = None

    e2 = Emulator(config=new, workdir=tmp_path / "resumed")
    out = e2.run()
    assert e2.modelname == "raven"
    assert isinstance(out.hydrograph, xr.Dataset)
    assert isinstance(out.storage, xr.Dataset)


def test_run_overwrite(gr4jcn_config, tmp_path):
    """Test that the emulator actually runs and returns the expected NSE."""
    gr4jcn, params = gr4jcn_config
    gr4jcn = gr4jcn.set_params(params)

    e = Emulator(config=gr4jcn, workdir=tmp_path)
    e.run()

    with pytest.raises(FileExistsError):
        e.run(overwrite=False)

    # Test that the emulator actually runs
    e.run(overwrite=True)


@pytest.mark.skip("Need to find a clean way to freeze emulator config instance.")
def test_emulator_config_is_read_only(dummy_config, tmp_path):
    cls, _ = dummy_config

    e = Emulator(config=cls(), workdir=tmp_path)

    # The emulator configuration should be read-only.
    with pytest.raises(TypeError):
        e.config.run_name = "Renamed"


def test_duplicate(gr4jcn_config, salmon_hru, tmp_path):
    hru = salmon_hru["land"]
    conf1 = GR4JCN(hrus=[hru], Duration=10)
    conf2 = conf1.duplicate(hrus=[hru])

    assert conf2.hrus == conf1.hrus
    assert conf2.duration == conf1.duration

    conf1.duration = 20
    assert conf2.duration == 10


def test_no_run_name(dummy_config, tmp_path):
    cls, P = dummy_config
    conf = cls(params=[0.5])

    paths = conf.write_rv(tmp_path, modelname="ham")
    assert "ham.rvi" in str(paths["rvi"])

    paths = conf.write_rv(tmp_path, overwrite=True)
    assert "raven.rvi" in str(paths["rvi"])


def test_set_params(gr4jcn_config):
    """Set params with numpy array, list and tuple."""
    gr4jcn, params = gr4jcn_config

    conf = gr4jcn.set_params(np.array([0.529, -3.396, 407.29, 1.072, 16.9, 0.947]))
    assert conf.params.GR4J_X1 == 0.529

    conf = gr4jcn.set_params([0.529, -3.396, 407.29, 1.072, 16.9, 0.947])
    assert conf.params.GR4J_X1 == 0.529

    conf = gr4jcn.set_params((0.529, -3.396, 407.29, 1.072, 16.9, 0.947))
    assert conf.params.GR4J_X1 == 0.529


def test_resume(gr4jcn_config, tmp_path):
    gr4jcn, params = gr4jcn_config
    gr4jcn = gr4jcn.set_params(params)

    # First run
    conf_a = gr4jcn.duplicate(RunName="a")
    a = Emulator(conf_a, workdir=tmp_path / "a")
    a.run()

    # Second run starting at the end of thr first one
    conf_b = a.resume()
    b = Emulator(conf_b, workdir=tmp_path / "b")
    out_b = b.run()

    # Complete run covering both periods
    conf_ab = gr4jcn.duplicate(Duration=gr4jcn.duration * 2, RunName="ab")
    ab = Emulator(conf_ab, workdir=tmp_path / "ab")
    out_ab = ab.run()

    for key in ["Soil Water[0]", "Soil Water[1]"]:
        np.testing.assert_array_almost_equal(out_b.storage[key] - out_ab.storage[key], 0, 5)

    # Confirm that the start_date is not changed when timestamp is False
    conf_c = a.resume(timestamp=False)
    assert conf_c.start_date == conf_a.start_date


# def test_version(self):
#     model = Raven()
#     assert model.raven_version == "3.6"
#
#     model = GR4JCN()
#     assert model.raven_version == "3.6"


def test_evaluation_periods(gr4jcn_config, tmp_path):
    """Test multiple evaluation periods are parsed correctly."""
    gr4jcn, params = gr4jcn_config

    evaluation_metrics = ["RMSE", "KLING_GUPTA", "KGE_PRIME"]

    conf = gr4jcn.set_params(params)
    conf.evaluation_metrics = evaluation_metrics
    conf.evaluation_period = [
        rc.EvaluationPeriod(name="period1", start="2000-01-01", end="2000-01-07"),
        rc.EvaluationPeriod(name="period2", start="2001-01-01", end="2000-01-15"),
    ]
    out = Emulator(conf, workdir=tmp_path).run()

    d = out.diagnostics
    for name in evaluation_metrics:
        assert f"DIAG_{name}" in d
        assert len(d[f"DIAG_{name}"]) == 3  # ALL, period1, period2


@pytest.mark.slow
@pytest.mark.online
@pytest.mark.xfail(error=OSError, reason="Network may be unreliable")
def test_run_with_dap_link(minimal_emulator, tmp_path):
    """Test Raven with DAP link instead of local netCDF file."""
    # Link to THREDDS Data Server netCDF testdata
    thredds = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/testdata/raven"
    dap_link = f"{thredds}/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"

    alt_names = {
        "RAINFALL": "rain",
        "TEMP_MIN": "tmin",
        "TEMP_MAX": "tmax",
        "PET": "pet",
        "HYDROGRAPH": "qobs",
        "SNOWFALL": "snow",
    }

    # Modifying of a session-scoped pytest fixture will cause problems with other tests.
    conf = minimal_emulator.model_copy(deep=True)
    conf.gauge = [rc.Gauge.from_nc(dap_link, alt_names=alt_names)]

    Emulator(conf, workdir=tmp_path).run()


def test_routing(yangtze):
    """We need at least two subbasins to activate routing."""
    from ravenpy.config.emulators.gr4jcn import P

    # Salmon catchment is now split into land- and lake-part.
    # The areas do not sum up to overall area of 4250.6 [km2].
    # This is the reason the "test_routing" will give different
    # results compared to "test_run". The "salmon_land_hru"
    # however, is kept in the overall area of 4250.6 [km2] such
    # that other tests still get the same results as before.

    salmon_land_hru_1 = dict(
        area=4250.6,
        elevation=843.0,
        latitude=54.4848,
        longitude=-123.3659,
        hru_type="land",
    )
    salmon_lake_hru_1 = dict(area=100.0, elevation=839.0, latitude=54.0, longitude=-123.4, hru_type="lake")
    salmon_land_hru_2 = dict(
        area=2000.0,
        elevation=835.0,
        latitude=54.123,
        longitude=-123.4234,
        hru_type="land",
    )

    salmon_river = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")
    ts_2d = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily_2d.nc")

    #########
    # R V P #
    #########

    parameters = [0.529, -3.396, 407.29, 1.072, 16.9, 0.947]

    #########
    # R V H #
    #########

    # Here we assume that we have two subbasins. The first one (subbasin_id=10)
    # has a lake (hru_id=2; area-100km2) and the rest is covered by land (hru_id=1;
    # area=4250.6km2). The second subbasin (subbasin_id=20) does not contain a
    # lake and is hence only land (hru_id=3; area=2000km2).
    #
    # Later the routing product will tell us which basin flows into which. Here
    # we assume that the first subbasin (subbasin_id=10) drains into the second
    # (subbasin_id=20). At the outlet of this second one we have an observation
    # station (see :ObservationData in RVT). We will compare these observations
    # with the simulated streamflow. That is the reason why "gauged=True" for
    # the second basin.

    # HRU IDs are 1 to 3
    # Sub-basin IDs are 10 and 20 (not 1 and 2), to help disambiguate
    hru1 = dict(hru_id=1, subbasin_id=10, **salmon_land_hru_1)
    hru2 = dict(hru_id=2, subbasin_id=10, **salmon_lake_hru_1)
    hru3 = dict(hru_id=3, subbasin_id=20, **salmon_land_hru_2)
    hrus = [hru1, hru2, hru3]

    hs = [
        rc.HRUState(hru_id=1, data={"SOIL[0]": P.GR4J_X1 * 1000 / 2, "SOIL[1]": 15}),
        rc.HRUState(hru_id=2, data={"SOIL[0]": 0, "SOIL[1]": 0}),
        rc.HRUState(hru_id=3, data={"SOIL[0]": P.GR4J_X1 * 1000 / 2, "SOIL[1]": 15}),
    ]

    # gauged = False:
    # Usually this output would only be written for user's convenience.
    # There is usually no observation of streamflow available within
    # catchments; only at the outlet. That's most commonly the reason
    # why a catchment is defined as it is defined.
    sub1 = dict(
        name="upstream",
        subbasin_id=10,
        downstream_id=20,
        profile="chn_10",
        gauged=False,
    )
    # gauged = True:
    # Since this is the outlet, this would usually be what we calibrate
    # against (i.e. we try to match this to Qobs).
    sub2 = dict(
        name="downstream",
        subbasin_id=20,
        downstream_id=-1,
        profile="chn_20",
        gauged=True,
    )

    # Set basin state variables to 0
    bs = [rc.BasinIndex(sb_id=10), rc.BasinIndex(sb_id=20)]

    SB_groups = [
        rc.SubBasinGroup(name="Land", sb_ids=[10]),
        rc.SubBasinGroup(name="Lakes", sb_ids=[20]),
    ]
    SBP = [
        rc.SBGroupPropertyMultiplier(group_name="Land", parameter_name="MANNINGS_N", mult=1.0),
        rc.SBGroupPropertyMultiplier(group_name="Lakes", parameter_name="RESERVOIR_CREST_WIDTH", mult=1.0),
    ]

    total_area_in_m2 = sum(hru["area"] for hru in hrus) * 1000 * 1000
    avg_annual_runoff = get_average_annual_runoff(ts_2d, total_area_in_m2)
    np.testing.assert_almost_equal(avg_annual_runoff, 139.5407534171111)

    r"""
    # These channel profiles describe the geometry of the actual river crossection.
    # The eight points (x) to describe the following geometry are given in each
    # profile:
    #
    # ----x                                     x---
    #      \           FLOODPLAIN             /
    #       x----x                     x----x
    #             \                  /
    #               \   RIVERBED   /
    #                 x----------x
    #
    """
    channel_profiles = [
        rc.ChannelProfile(
            name="chn_10",
            bed_slope=7.62066e-05,
            survey_points=[
                (0, 463.647),
                (16.0, 459.647),
                (90.9828, 459.647),
                (92.9828, 458.647),
                (126.4742, 458.647),
                (128.4742, 459.647),
                (203.457, 459.647),
                (219.457, 463.647),
            ],
            roughness_zones=[
                (0, 0.0909167),
                (90.9828, 0.035),
                (128.4742, 0.0909167),
            ],
        ),
        rc.ChannelProfile(
            name="chn_20",
            bed_slope=9.95895e-05,
            survey_points=[
                (0, 450.657),
                (16.0, 446.657),
                (85.0166, 446.657),
                (87.0166, 445.657),
                (117.5249, 445.657),
                (119.5249, 446.657),
                (188.54149999999998, 446.657),
                (204.54149999999998, 450.657),
            ],
            roughness_zones=[
                (0, 0.0915769),
                (85.0166, 0.035),
                (119.5249, 0.0915769),
            ],
        ),
    ]
    #########
    # R V T #
    #########

    alt_names = {
        "RAINFALL": "rain",
        "TEMP_MIN": "tmin",
        "TEMP_MAX": "tmax",
        "SNOWFALL": "snow",
        "PET": "pet",
    }

    data_type = ["RAINFALL", "TEMP_MIN", "TEMP_MAX", "SNOWFALL", "PET"]
    gw = rc.GridWeights(
        number_hrus=3,
        number_grid_cells=1,
        # Here we have a special case: station is 0 for every row because the example NC
        # has only one region/station (which is column 0)
        data=((1, 0, 1.0), (2, 0, 1.0), (3, 0, 1.0)),
    )

    sf = [rc.StationForcing.from_nc(ts_2d, typ, alt_names=alt_names[typ], GridWeights=gw) for typ in data_type]

    obs = [rc.ObservationData.from_nc(salmon_river, alt_names="qobs", uid=20)]

    ###################
    # Configure model #
    ###################

    model = GR4JCN(
        StartDate=dt.datetime(2000, 1, 1),
        EndDate=dt.datetime(2002, 1, 1),
        RunName="test_gr4jcn_routing",
        Routing="ROUTE_DIFFUSIVE_WAVE",
        params=parameters,
        HRUs=hrus,
        HRUStateVariableTable=hs,
        SubBasins=[sub1, sub2],
        SubBasinGroup=SB_groups,
        SBGroupPropertyMultiplier=SBP,
        BasinStateVariables=bs,
        ChannelProfile=channel_profiles,
        StationForcing=sf,
        ObservationData=obs,
        GlobalParameter={"AVG_ANNUAL_RUNOFF": avg_annual_runoff},
        WriteForcingFunctions=True,
        UniformInitialConditions=None,
        EvaluationMetrics=("NASH_SUTCLIFFE", "KGE_PRIME"),
    )

    #############
    # Run model #
    #############

    out = Emulator(model).run()

    ###########
    # Verify  #
    ###########

    hds = out.hydrograph.q_sim

    assert len(hds.nbasins) == 1  # the number of "gauged" basins is 1

    # We only have one SB with gauged=True, so the output has a single column.
    # The number of time steps simulated between (2000, 1, 1) and
    # (2002, 1, 1) is 732.
    assert hds.shape == (732, 1)

    # Check simulated streamflow at first three timesteps and three simulated
    # timesteps in the middle of the simulation period.
    dates = (
        "2000-01-01",
        "2000-01-02",
        "2000-01-03",
        "2001-01-30",
        "2001-01-31",
        "2001-02-01",
    )

    target_q_sim = [0.0, 0.304073, 0.980807, 17.54049, 17.409493, 17.437954]

    for t in range(6):
        np.testing.assert_almost_equal(hds.sel(nbasins=0, time=dates[t]), target_q_sim[t], 4)

    # For lumped GR4J model we have 1 subbasin and 1 HRU as well as no routing, no
    # channel profiles, and the area of the entire basin is 4250.6 [km2]. Comparison
    # of simulated and observed streamflow at outlet yielded:
    # np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.116971, 4)
    #
    # This is now a different value due to:
    # - basin we have here is larger (4250.6 [km2] + 100 [km2] + 2000.0 [km2])
    # - we do routing: so water from subbasin 1 needs some time to arrive at the
    #   outlet of subbasin 2
    d = out.diagnostics
    np.testing.assert_almost_equal(d["DIAG_NASH_SUTCLIFFE"], -0.0141168, 4)

    assert len(list(out.path.glob("*ForcingFunctions.nc"))) == 1


@pytest.mark.slow
@pytest.mark.xfail
def test_routing_lievre_tutorial(tmp_path, yangtze):
    from ravenpy.extractors.routing_product import (
        BasinMakerExtractor,
        GridWeightExtractor,
        open_shapefile,
    )

    ###############
    # Input files #
    ###############

    routing_product_shp_path = yangtze.fetch("raven-routing-sample/lievre_hrus_v21.zip")

    vic_streaminputs_nc_path = yangtze.fetch("raven-routing-sample/VIC_streaminputs.nc")
    vic_temperatures_nc_path = yangtze.fetch("raven-routing-sample/VIC_temperatures.nc")

    observation_data_nc_path = yangtze.fetch("raven-routing-sample/WSC02LE024.nc")

    streaminputs = xr.open_dataset(vic_streaminputs_nc_path)

    attrs = {
        "StartDate": streaminputs.indexes["time"][0],
        "EndDate": streaminputs.indexes["time"][-4],
        "TimeStep": "24:00:00",
        "EvaluationMetrics": [
            "NASH_SUTCLIFFE",
            "PCT_BIAS",
            "KLING_GUPTA",
        ],
    }

    #######
    # RVH #
    #######

    rvh_extractor = BasinMakerExtractor(
        open_shapefile(routing_product_shp_path),
        hru_aspect_convention="ArcGIS",
    )
    rvh_config = rvh_extractor.extract()

    # There are 3 gauged subbasins in the HRU file generated for the Routing Product V2.1,
    # but we are only interested in the most downstream one, "02LE024".

    gauged_sbs = [sb for sb in rvh_config["sub_basins"] if sb["gauged"]]
    assert len(gauged_sbs) == 3

    gauged_sb = None  # we're looking for gauge 02LE024 which is the most downstream
    for sb in gauged_sbs:
        if sb["gauge_id"] == "02LE024":
            gauged_sb = sb
        else:
            # Raven will ignore the other two
            sb["gauged"] = False
            sb["gauge_id"] = ""

    attrs.update(rvh_config)
    attrs["SBGroupPropertyMultiplier"] = [
        {"group_name": "Land", "parameter_name": "MANNINGS_N", "mult": 1.0},
        {"group_name": "Lakes", "parameter_name": "RESERVOIR_CREST_WIDTH", "mult": 1.0},
    ]
    attrs["GlobalParameter"] = {"AVG_ANNUAL_RUNOFF": 597.6287}
    attrs["SoilClasses"] = [dict(name="AQUIFER")]
    attrs["SoilProfiles"] = [
        dict(name="LAKE", soil_classes=["AQUIFER"], thicknesses=[5]),
        dict(name="Soil_Land_HRU", soil_classes=["AQUIFER"], thicknesses=[5]),
    ]
    attrs["VegetationClasses"] = [
        dict(name="Veg_Land_HRU", max_ht=25, max_lai=5, max_leaf_cond=5),
        dict(
            name="WATER",
        ),
    ]
    attrs["LandUseClasses"] = [
        dict(name="Landuse_Land_HRU", impermeable_frac=0, forest_coverage=1),
        dict(
            name="WATER",
        ),
    ]

    gw_pr = GridWeightExtractor(
        vic_streaminputs_nc_path,
        routing_product_shp_path,
        var_names=["lon", "lat"],
        routing_id_field="HRU_ID",
    ).extract()
    gw_pr_f = tmp_path / "gw_pr.rvt"
    gw_pr_f.write_text(str(rc.GridWeights(**gw_pr)))

    gf_pr = rc.GriddedForcing.from_nc(
        vic_streaminputs_nc_path,
        data_type="PRECIP",
        alt_names=("Streaminputs",),
        grid_weights=gw_pr,
        dim_names_nc=("lon_dim", "lat_dim", "time"),
    )

    assert gf_pr.linear_transform.scale == 4
    assert gf_pr.linear_transform.offset == 0

    gw_tas = GridWeightExtractor(
        vic_temperatures_nc_path,
        routing_product_shp_path,
        var_names=["lon", "lat"],
        routing_id_field="HRU_ID",
    ).extract()
    gw_tas_f = tmp_path / "gw_tas.rvt"
    gw_tas_f.write_text(str(rc.GridWeights(**gw_tas)))

    gf_tas = rc.GriddedForcing.from_nc(
        vic_temperatures_nc_path,
        data_type="TEMP_AVE",
        alt_names=("Avg_temp",),
        grid_weights=gw_tas,
        dim_names_nc=("lon_dim", "lat_dim", "time"),
    )

    obs = rc.ObservationData.from_nc(observation_data_nc_path, uid=gauged_sb["subbasin_id"], alt_names=("Q",))

    conf = BasicRoute(**attrs, GriddedForcing=[gf_tas, gf_pr], ObservationData=[obs])

    out = Emulator(conf, workdir=tmp_path).run()

    ##########
    # Verify #
    ##########

    assert out.hydrograph.basin_name.item() == gauged_sb["name"]

    csv_lines = out.files["diagnostics"].read_text().split("\n")

    assert csv_lines[1].split(",")[:-1] == [
        "HYDROGRAPH_ALL[3077541]",
        str(observation_data_nc_path),
        "0.253959",  # NASH_SUTCLIFFE "0.253959",
        "-17.0904",  # PCT_BIAS "-17.0904"
        "0.443212",  # KLING_GUPTA "0.443212"
    ]

    for d, q_sim in [
        (0, 85.92355875229545),
        (1000, 74.05569855818379),
        (2000, 62.675159400333115),
        (3000, 42.73584909530037),
        (4000, 128.70284018326998),
    ]:
        assert out.hydrograph.q_sim[d].item() == pytest.approx(q_sim)


@pytest.mark.online
def test_canopex():
    CANOPEX_DAP = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/ets/Watersheds_5797_cfcompliant.nc"

    # Set HRU
    hru = dict(
        area=3650.47,
        elevation=330.59,
        latitude=49.51,
        longitude=-95.72,
        hru_type="land",
    )
    data_type = ["TEMP_MAX", "TEMP_MIN", "PRECIP"]

    alt_names = {
        "TEMP_MIN": "tasmin",
        "TEMP_MAX": "tasmax",
        "PRECIP": "pr",
    }

    data_kwds = {
        "ALL": {
            "elevation": hru["elevation"],
            "latitude": hru["latitude"],
            "longitude": hru["longitude"],
        }
    }

    basin = 5600

    qobs = [rc.ObservationData.from_nc(CANOPEX_DAP, alt_names="discharge", station_idx=basin, engine="netcdf4")]

    gauge_data = [
        rc.Gauge.from_nc(
            CANOPEX_DAP,
            station_idx=basin,
            data_type=data_type,  # Note that this is the list of all the variables
            alt_names=alt_names,  # Note that all variables here are mapped to their names in the netcdf file.
            engine="netcdf4",
            data_kwds=data_kwds,
        )
    ]

    # Set config
    model = GR4JCN(
        start_date=dt.datetime(2010, 6, 1),
        end_date=dt.datetime(2010, 6, 10),
        Gauge=gauge_data,
        ObservationData=qobs,
        run_name="Test_run",
        rain_snow_fraction="RAINSNOW_DINGMAN",
        hrus=[hru],
        params=[108.02, 2.8693, 25.352, 1.3696, 1.2483, 0.30679],
    )

    # Test that it at least runs while we're here
    Emulator(config=model, overwrite=True).run(overwrite=True)

    # Check unit transformation parameters are correctly inferred
    var = model.gauge[0].data[0].data_type
    assert var in ["TEMP_MIN", "TEMP_MAX", "PRECIP"]

    if var in ["TEMP_MIN", "TEMP_MAX"]:
        assert model.gauge[0].data[0].read_from_netcdf.linear_transform.scale == 1.0
        assert model.gauge[0].data[0].read_from_netcdf.linear_transform.offset == -273.15
    else:
        assert model.gauge[0].data[0].read_from_netcdf.linear_transform.scale == 86400.0
        assert model.gauge[0].data[0].read_from_netcdf.linear_transform.offset == 0.0
